#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Some tools that generate VCF (callers) will result in very slightly different qual
scores and other floating-point-valued fields when run on different hardware. This
handler ignores the QUAL and INFO columns and only compares the genotype (GT) field
of sample columns. Only works for single-sample VCFs.
"""
from functools import partial
from pathlib import Path
import re

import subby

from pytest_wdl.data_types import DataFile, assert_text_files_equal, diff_default
from pytest_wdl.utils import tempdir


GENO_RE = re.compile("[|/]")


class VcfDataFile(DataFile):
    def _assert_contents_equal(self, other_path: Path, other_opts: dict) -> None:
        compare_phase = (
            self.compare_opts.get("compare_phase") or
            other_opts.get("compare_phase")
        )
        try:
            assert_text_files_equal(
                self.path,
                other_path,
                self._get_allowed_diff_lines(other_opts),
                diff_fn=partial(diff_vcf_columns, compare_phase=compare_phase)
            )
        except AssertionError as err:
            raise AssertionError(
                f"VCF files are not equal: {self.path} != {other_path}"
            ) from err


def diff_vcf_columns(file1: Path, file2: Path, compare_phase: bool = False) -> int:
    with tempdir() as temp:
        def make_comparable(infile, outfile):
            cmd = ["grep -vE '^#'", "cut -f 1-5,7,10", "cut -d ':' -f 1"]
            output = subby.sub(cmd, stdin=infile)
            with open(outfile, "wt") as out:
                if compare_phase:
                    out.write(output)
                else:
                    # Normalize the allele separator and sort the alleles
                    for row in output.splitlines(keepends=True):
                        r, g = row.rsplit("\t", 1)
                        g_strip = g.rstrip()
                        g_norm = "/".join(sorted(GENO_RE.split(g_strip)))
                        out.write(f"{r}\t{g_norm}")
                        if len(g) != len(g_strip):
                            out.write("\n")

        cmp_file1 = temp / "file1"
        cmp_file2 = temp / "file2"
        make_comparable(file1, cmp_file1)
        make_comparable(file2, cmp_file2)
        return diff_default(cmp_file1, cmp_file2)



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import time

import subby
import pytest

from pytest_wdl.executors import ENV_JAVA_HOME
from pytest_wdl.executors.cromwell_local import ENV_CROMWELL_JAR


@pytest.mark.integration
@pytest.mark.remote
def test_cromwell_server_workflow(user_config, workflow_data, workflow_runner):
    cromwell_jar_file = user_config.get_executor_defaults("cromwell").get(
        "cromwell_jar_file", os.environ.get(ENV_CROMWELL_JAR)
    )

    java_jar = user_config.get_executor_defaults("cromwell").get(
        "java_bin", os.environ.get(ENV_JAVA_HOME) + "/bin/java"
    )

    p = subby.run(
        f"{java_jar} -jar {cromwell_jar_file} server | tee /dev/stderr", block=False
    )
    time.sleep(10)
    inputs = {
        "in_txt": workflow_data["in_txt"],
        "in_int": 1
    }
    outputs = {
        "out_txt": workflow_data["out_txt"],
        "out_int": 1
    }

    try:
        workflow_runner(
            "test.wdl",
            inputs,
            outputs,
            executors=["cromwell-server"]
        )
    finally:
        p.kill()



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json
import os
from pathlib import Path
import tempfile
from typing import Optional, Sequence, Union

import subby

from pytest_wdl.executors import (
    ExecutionFailedError,
    JavaExecutor,
    get_target_name,
    read_write_inputs,
)
from pytest_wdl.executors._cromwell import (
    ENV_CROMWELL_ARGS, ENV_CROMWELL_JAR, ENV_CROMWELL_CONFIG, CromwellHelperMixin
)
from pytest_wdl.utils import LOG, ensure_path


class CromwellLocalExecutor(JavaExecutor, CromwellHelperMixin):
    """
    Manages the running of WDL workflows using Cromwell.

    Args:
        import_dirs: Relative or absolute paths to directories containing WDL
            scripts that should be available as imports.
        java_bin: Path to the java executable.
        java_args: Default Java arguments to use; can be overidden by passing
            `java_args=...` to `run_workflow`.
        cromwell_jar_file: Path to the Cromwell JAR file.
        cromwell_args: Default Cromwell arguments to use; can be overridden by
            passing `cromwell_args=...` to `run_workflow`.
    """

    def __init__(
        self,
        import_dirs: Optional[Sequence[Path]] = None,
        java_bin: Optional[Union[str, Path]] = None,
        java_args: Optional[str] = None,
        cromwell_jar_file: Optional[Union[str, Path]] = None,
        cromwell_configuration: Optional[Union[str, Path, dict]] = None,
        cromwell_args: Optional[str] = None,
        # deprecated
        cromwell_config_file: Optional[Union[str, Path]] = None,
    ):
        super().__init__(java_bin, java_args)

        self._import_dirs = import_dirs

        self._cromwell_jar_file = self.resolve_jar_file(
            "cromwell*.jar", cromwell_jar_file, ENV_CROMWELL_JAR
        )

        if cromwell_config_file:
            LOG.warn(
                "The 'cromwell_config_file' parameter is deprecated; please use "
                "'cromwell_configuration' instead."
            )
            if not cromwell_configuration:
                cromwell_configuration = cromwell_config_file

        if not cromwell_configuration:
            config_file = os.environ.get(ENV_CROMWELL_CONFIG)

            if config_file:
                cromwell_configuration = ensure_path(config_file)

        if cromwell_configuration:
            if self.java_args:
                LOG.warn("'cromwell_configuration' is ignored when 'java_args' are set")
            else:
                if isinstance(cromwell_configuration, dict):
                    cromwell_config_file = Path(tempfile.mkstemp(suffix=".zip")[1])
                    with open(cromwell_config_file, "wt") as out:
                        json.dump(cromwell_configuration, out)
                else:
                    cromwell_config_file = ensure_path(
                        cromwell_configuration, is_file=True, exists=True
                    )

                self.java_args = f"-Dconfig.file={cromwell_config_file}"

        self._cromwell_args = cromwell_args or os.environ.get(ENV_CROMWELL_ARGS)

    def run_workflow(
        self,
        wdl_path: Path,
        inputs: Optional[dict] = None,
        expected: Optional[dict] = None,
        **kwargs,
    ) -> dict:
        """
        Run a WDL workflow on given inputs, and check that the output matches
        given expected values.

        Args:
            wdl_path: The WDL script to execute.
            inputs: Object that will be serialized to JSON and provided to Cromwell
                as the workflow inputs.
            expected: Dict mapping output parameter names to expected values.
            kwargs: Additional keyword arguments, mostly for debugging:
                * workflow_name: The name of the workflow in the WDL script. If None,
                    the name of the WDL script is used (without the .wdl extension).
                * inputs_file: Path to the Cromwell inputs file to use. Inputs are
                    written to this file only if it doesn't exist.
                * imports_file: Path to the WDL imports file to use. Imports are
                    written to this file only if it doesn't exist.
                * java_args: Additional arguments to pass to Java runtime.
                * cromwell_args: Additional arguments to pass to `cromwell run`.

        Returns:
            Dict of outputs.

        Raises:
            ExecutionFailedError: if there was an error executing Cromwell
            AssertionError: if the actual outputs don't match the expected outputs
        """
        target, is_task = get_target_name(
            wdl_path=wdl_path, import_dirs=self._import_dirs, **kwargs
        )

        if is_task:
            raise ValueError(
                "Cromwell cannot execute tasks independently of a workflow"
            )

        inputs_dict, inputs_file = read_write_inputs(
            inputs_file=kwargs.get("inputs_file"), inputs_dict=inputs, namespace=target
        )
        imports_file = self._get_workflow_imports(
            self._import_dirs, kwargs.get("imports_file")
        )

        inputs_arg = f"-i {inputs_file}" if inputs_file else ""
        imports_zip_arg = f"-p {imports_file}" if imports_file else ""
        java_args = kwargs.get("java_args", self.java_args) or ""
        cromwell_args = kwargs.get("cromwell_args", self._cromwell_args) or ""
        metadata_file = Path.cwd() / "metadata.json"

        cmd = (
            f"{self.java_bin} {java_args} -jar {self._cromwell_jar_file} run "
            f"-m {metadata_file} {cromwell_args} {inputs_arg} {imports_zip_arg} "
            f"{wdl_path}"
        )
        LOG.info(
            f"Executing cromwell command '{cmd}' with inputs "
            f"{json.dumps(inputs_dict, default=str)}"
        )

        exe = subby.run(cmd, raise_on_error=False)

        metadata = None

        if metadata_file.exists():
            with open(metadata_file, "rt") as inp:
                metadata = json.load(inp)

        if exe.ok:
            if metadata:
                assert metadata["status"] == "Succeeded"
                outputs = metadata["outputs"]
            else:
                LOG.warning(
                    f"Cromwell command completed successfully but did not generate "
                    f"a metadata file at {metadata_file}"
                )
                outputs = self._get_cromwell_outputs(exe.output)
        else:
            error_kwargs = {
                "executor": "cromwell",
                "target": target,
                "status": "Failed",
                "inputs": inputs_dict,
                "executor_stdout": exe.output,
                "executor_stderr": exe.error,
            }
            if metadata:
                error_kwargs = {
                    "executor": "cromwell",
                    "target": target,
                    "status": "Failed",
                    "inputs": inputs_dict,
                }
                self._parse_metadata_errors(
                    metadata, target=target, error_kwargs=error_kwargs
                )
            else:
                error_kwargs["msg"] = (
                    f"Cromwell command failed but did not generate a metadata "
                    f"file at {metadata_file}"
                )

            raise ExecutionFailedError(**error_kwargs)

        if expected:
            self._validate_outputs(outputs, expected, target)

        return outputs



import contextlib
from pathlib import Path
import random
import string
from typing import cast
from urllib.request import urlopen

import pytest
from _pytest.fixtures import FixtureRequest

from pytest_wdl.executors import parse_wdl
from pytest_wdl.plugins import PluginError
from pytest_wdl.url_schemes import Response
from pytest_wdl.utils import tempdir

SKIP_REASON = "dxpy is not installed; DNAnexus URL handler and dxWDL executor will " \
              "not be tested"

try:
    dx = pytest.importorskip("pytest_wdl.providers.dx", reason=SKIP_REASON)
    dxpy = dx.dxpy
    # Force non-interactive login since we may be running in a CI/CD environment
    with dx.login(interactive=False):
        assert dxpy.whoami()
except PluginError as err:
    dxpy = None

pytestmark = pytest.mark.skipif(dxpy is None, reason=SKIP_REASON)


DX_FILE_ID = "file-BgY4VzQ0bvyg22pfZQpXfzgK"
DX_PROJECT_ID = "project-BQbJpBj0bvygyQxgQ1800Jkk"


@contextlib.contextmanager
def random_project_folder(length: int = 8, prefix: str = "") -> str:
    """
    ContextManager that generates a random folder name, ensures that it doesn't
    already exist in the current project, yields it, and then deletes the folder if it
    exists.

    Returns:
        The folder path.
    """
    letters = string.ascii_letters + string.digits
    project = dxpy.DXProject(dxpy.PROJECT_CONTEXT_ID)

    while True:
        folder = "".join(random.choices(letters, k=length))
        folder_path = f"{prefix}{'' if prefix.endswith('/') else '/'}{folder}"
        try:
            project.list_folder(folder_path)
        except dxpy.exceptions.ResourceNotFound:
            # We found a folder that does not exist
            break

    try:
        yield folder_path
    finally:
        project.remove_folder(folder_path, recurse=True, force=True)


def test_dx_scheme():
    rsp = urlopen(f"dx://{DX_FILE_ID}")
    assert isinstance(rsp, Response)
    import pytest_wdl.providers.dx
    assert isinstance(rsp, pytest_wdl.providers.dx.DxResponse)
    with tempdir() as d:
        outfile = d / "readme.txt"
        cast(Response, rsp).download_file(outfile, False)
        assert outfile.exists()
        with open(outfile, "rt") as inp:
            txt = inp.read()
        assert txt.startswith("README.1st.txt")
        assert txt.rstrip().endswith(
            "SRR100022: Full exome to use as input to your analyses."
        )


def test_dx_scheme_with_project():
    rsp = urlopen(f"dx://{DX_PROJECT_ID}:{DX_FILE_ID}")
    assert isinstance(rsp, Response)
    import pytest_wdl.providers.dx
    assert isinstance(rsp, pytest_wdl.providers.dx.DxResponse)
    with tempdir() as d:
        outfile = d / "readme.txt"
        cast(Response, rsp).download_file(outfile, False)
        assert outfile.exists()
        with open(outfile, "rt") as inp:
            txt = inp.read()
        assert txt.startswith("README.1st.txt")
        assert txt.rstrip().endswith(
            "SRR100022: Full exome to use as input to your analyses."
        )


def test_dx_input_formatter(request: FixtureRequest):
    doc = parse_wdl(Path(request.fspath.dirpath()) / "test_dx.wdl")
    dx_input_formatter = dx.DxInputsFormatter(doc)

    assert dx_input_formatter.format_inputs({
        "str": "abc",
        "array_str": ["def", "ghi"],
        "map_str_str": {
            "jkl": "mno"
        },
        "struc": {
            "str": "pqr",
            "array_str": ["stu", "vwx"],
            "map_str_str": {
                "xy1": "234"
            }
        },
        "array_struc": [{
            "str": "567",
            "array_str": ["890", "ABC"],
            "map_str_str": {
                "DEF": "GHI"
            }
        }],
        "map_str_struc": {
            "JKL": {
                "str": "MNO",
                "array_str": ["PQR", "STU"],
                "map_str_str": {
                    "VWX": "YZ1"
                }
            }
        },
        "pair_str_str": {
            "left": "A",
            "right": "B"
        },
        "array_pair_str_str": [
            {
                "left": "X",
                "right": "Y"
            }
        ]
    }) == {
        "str": "abc",
        "array_str": ["def", "ghi"],
        "map_str_str": {
            "___": {
                "keys": ["jkl"],
                "values": ["mno"]
            }
        },
        "struc": {
            "___": {
                "str": "pqr",
                "array_str": ["stu", "vwx"],
                "map_str_str": {
                    "keys": ["xy1"],
                    "values": ["234"]
                }
            }
        },
        "array_struc": {
            "___": [{
                "str": "567",
                "array_str": ["890", "ABC"],
                "map_str_str": {
                    "keys": ["DEF"],
                    "values": ["GHI"]
                }
            }]
        },
        "map_str_struc": {
            "___": {
                "keys": ["JKL"],
                "values": [{
                    "str": "MNO",
                    "array_str": ["PQR", "STU"],
                    "map_str_str": {
                        "keys": ["VWX"],
                        "values": ["YZ1"]
                    }
                }]
            }
        },
        "pair_str_str": {
            "___": {
                "left": "A",
                "right": "B"
            }
        },
        "array_pair_str_str": {
            "___": [
                {
                    "left": "X",
                    "right": "Y"
                }

            ]
        }
    }


def test_dx_input_formatter_with_data_files():
    """TODO"""


@pytest.mark.integration
@pytest.mark.remote
def test_dxwdl_workflow(workflow_data, workflow_runner):
    with random_project_folder() as workflow_folder:
        inputs = {
            "in_txt": workflow_data["in_txt"],
            "in_int": 1
        }
        outputs = {
            "out_txt": workflow_data["out_txt"],
            "out_int": 1
        }
        workflow_runner(
            "test.wdl",
            inputs,
            outputs,
            executors=["dxwdl"],
            workflow_folder=workflow_folder
        )


# TODO: implement task support for dxWDL executor
# @pytest.mark.integration
# @pytest.mark.remote
# def test_dxwdl_task(workflow_data, workflow_runner):
#     with random_project_folder() as workflow_folder:
#         inputs = {
#             "in_txt": workflow_data["in_txt"],
#             "in_int": 1
#         }
#         outputs = {
#             "out_txt": workflow_data["out_txt"],
#             "out_int": 1
#         }
#         workflow_runner(
#             "test.wdl",
#             inputs,
#             outputs,
#             executors=["dxwdl"],
#             task_name="cat",
#             workflow_folder=workflow_folder
#         )



import textwrap

from pytest_wdl.executors import ExecutionFailedError


def test_execution_failed_error():
    try:
        raise ExecutionFailedError(
            executor="foo",
            target="target",
            status="Failed",
            inputs={
                "a": 1,
                "b": "bar"
            },
            executor_stdout="dis happen",
            executor_stderr="dis an err\noopsy",
            failed_task="baz",
            failed_task_exit_status=1,
            failed_task_stdout="oh oh\nsomeone set us up the bomb",
            failed_task_stderr="beep boop beep\nI am except"
        )
    except ExecutionFailedError as err:
        assert str(err) == textwrap.dedent(f"""
        foo failed with status Failed while running task baz of target:
            inputs:
                {{'a': 1, 'b': 'bar'}}
            executor_stdout:
                dis happen
            executor_stderr:
                dis an err
                oopsy
            failed_task_exit_status: 1
            failed_task_stdout:
                oh oh
                someone set us up the bomb
            failed_task_stderr:
                beep boop beep
                I am except
        """)



import json
from pathlib import Path

from pytest_wdl.data_types import DataFile


class JsonDataFile(DataFile):
    def _assert_contents_equal(self, other_path: Path, other_opts: dict) -> None:
        with open(self.path, "rt") as inp:
            try:
                j1 = json.load(inp)
            except json.decoder.JSONDecodeError:
                raise AssertionError(f"Invalid JSON file {self.path}")
        with open(other_path, "rt") as inp:
            try:
                j2 = json.load(inp)
            except json.decoder.JSONDecodeError:
                raise AssertionError(f"Invalid JSON file {other_path}")
        assert j1 == j2



from unittest.mock import Mock

import pytest

from pytest_wdl.plugins import plugin_factory_map


def test_plugin_factory_map():
    ep1 = Mock()
    ep1.name = "foo"
    ep1.module_name = "pytest_wdl.foo"
    ep2 = Mock()
    ep2.name = "foo"
    ep2.module_name = "bar.baz"
    entry_points = [ep1, ep2]
    pfmap = plugin_factory_map(None, entry_points=entry_points)
    assert len(pfmap) == 1
    assert "foo" in pfmap
    assert pfmap["foo"].entry_point == ep2

    ep3 = Mock()
    ep3.name = "foo"
    ep3.module_name = "blorf.bleep"
    entry_points.append(ep3)
    with pytest.raises(RuntimeError):
        plugin_factory_map(None, entry_points=entry_points)



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from abc import ABCMeta, abstractmethod
from enum import Enum
import functools
from pathlib import Path
from typing import Optional, Sequence
from urllib.request import BaseHandler, Request, build_opener, install_opener

from pkg_resources import iter_entry_points

from pytest_wdl.plugins import PluginError, PluginFactory
from pytest_wdl.utils import LOG, verify_digests

try:
    from tqdm import tqdm as progress
except ImportError:  # pragma: no-cover
    LOG.debug(
        "tqdm is not installed; progress bar will not be displayed when "
        "downloading files"
    )
    progress = None


class Method(Enum):
    OPEN = ("urlopen", "{}_open")
    REQUEST = ("request", "{}_request")
    RESPONSE = ("response", "{}_response")

    def __init__(self, src_attr, dest_pattern):
        self.src_attr = src_attr
        self.dest_pattern = dest_pattern


class Response(metaclass=ABCMeta):
    @abstractmethod
    def download_file(
        self,
        destination: Path,
        show_progress: bool = False,
        digests: Optional[dict] = None
    ):
        """
        Download a file to a specific destination.

        Args:
            destination: Destination path
            show_progress: Whether to show a progress bar
            digests: Optional dict mapping hash names to digests. These are used to
                validate the downloaded file.

        Raises:
            DigestsNotEqualError
        """
        pass


class BaseResponse(Response, metaclass=ABCMeta):
    @abstractmethod
    def get_content_length(self) -> Optional[int]:
        pass

    @abstractmethod
    def read(self, block_size: int):
        pass

    def download_file(
        self,
        destination: Path,
        show_progress: bool = False,
        digests: Optional[dict] = None
    ):
        total_size = self.get_content_length()
        block_size = 16 * 1024
        if total_size and total_size < block_size:
            block_size = total_size

        if show_progress and progress:
            progress_bar = progress(
                total=total_size,
                unit="b",
                unit_scale=True,
                unit_divisor=1024,
                desc=f"Localizing {destination.name}"
            )

            def progress_reader():
                b = self.read(block_size)
                if b:
                    progress_bar.update(block_size)
                else:
                    progress_bar.close()
                return b

            reader = progress_reader
        else:
            reader = functools.partial(self.read, block_size)

        downloaded_size = 0

        with open(destination, "wb") as out:
            while True:
                buf = reader()
                if not buf:
                    break
                downloaded_size += len(buf)
                out.write(buf)

        if downloaded_size != total_size:  # TODO: test this
            raise AssertionError(
                f"Size of downloaded file {destination} does not match expected size "
                f"{total_size}"
            )

        if digests:
            verify_digests(destination, digests)


class ResponseWrapper(BaseResponse):
    def __init__(self, rsp):
        self.rsp = rsp

    def get_content_length(self) -> Optional[int]:
        size_str = self.rsp.getheader("content-length")
        if size_str:
            return int(size_str)

    def read(self, block_size: int) -> bytes:
        return self.rsp.read(block_size)


class UrlHandler(BaseHandler, metaclass=ABCMeta):
    @property
    @abstractmethod
    def scheme(self) -> str:
        pass

    @property
    def handles(self) -> Sequence[Method]:
        return []  # pragma: no-cover

    def alias(self):
        """
        Add aliases that are required by urllib for handled methods.
        """
        for method in self.handles:
            src = getattr(self, method.src_attr)
            setattr(self, method.dest_pattern.format(self.scheme), src)

    def request(self, request: Request) -> Request:
        pass

    def urlopen(self, request: Request) -> Response:
        pass

    def response(self, request: Request, response: Response) -> Response:
        pass


def install_schemes():
    def create_handler(_entry_point):
        handler_factory = PluginFactory(_entry_point, UrlHandler)
        handler = handler_factory()
        handler.alias()
        return handler

    handlers = []

    for entry_point in iter_entry_points(group="pytest_wdl.url_schemes"):
        try:
            handlers.append(create_handler(entry_point))
        except PluginError:
            pass

    install_opener(build_opener(*handlers))



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Convert BAM to SAM for diff.
"""
from enum import Enum
from functools import partial
from pathlib import Path
import re
from typing import Iterable, Optional

import subby

from pytest_wdl.data_types import DataFile, assert_text_files_equal, diff_default
from pytest_wdl.utils import tempdir

# TODO: fall back to command line samtools (if installed)
import pysam


INVARIATE_COLUMNS = "1,2,5,10,11"
ALL_COLUMNS = "1-11"


class Sorting(Enum):
    NONE = 0
    COORDINATE = 1
    NAME = 2


class BamDataFile(DataFile):
    """
    Supports comparing output of BAM file. This uses pysam to convert BAM to
    SAM, so that DataFile can carry out a regular diff on the SAM files.
    """
    def _assert_contents_equal(self, other_path: Path, other_opts: dict):
        try:
            assert_bam_files_equal(
                self.path,
                other_path,
                allowed_diff_lines=self._get_allowed_diff_lines(other_opts),
                min_mapq=self._get_min_mapq(other_opts),
                compare_tag_columns=self._get_compare_tag_columns(other_opts)
            )
        except AssertionError as err:
            raise AssertionError(
                f"BAM files are not equal: {self.path} != {other_path}"
            ) from err

    def _get_min_mapq(self, other_opts: dict) -> int:
        return max(
            self.compare_opts.get("min_mapq", 0),
            other_opts.get("min_mapq", 0)
        )

    def _get_compare_tag_columns(self, other_opts: dict) -> bool:
        return (
            self.compare_opts.get("compare_tag_columns") or
            other_opts.get("compare_tag_columns")
        )


def assert_bam_files_equal(
    file1: Path,
    file2: Path,
    allowed_diff_lines: int = 0,
    min_mapq: int = 0,
    compare_tag_columns: bool = False
):
    """
    Compare two BAM files:
    * Convert them to SAM format
    * Optionally re-sort the files by chromosome, position, and flag
    * First compare all lines using only a subset of columns that should be
    deterministic
    * Next, filter the files by MAPQ and compare the remaining rows using all columns

    Args:
        file1: First BAM to compare
        file2: Second BAM to compare
        allowed_diff_lines: Number of lines by which the BAMs are allowed to differ
            (after being convert to SAM)
        min_mapq: Minimum mapq used to filter reads when comparing all columns
        compare_tag_columns: Whether to include tag columns (12+) when comparing
            all columns
    """
    with tempdir() as temp:
        # Compare all reads using subset of columns
        cmp_file1 = temp / "all_reads_subset_columns_file1"
        cmp_file2 = temp / "all_reads_subset_columns_file2"
        bam_to_sam(
            file1,
            cmp_file1,
            headers=None,
            sorting=Sorting.NAME
        )
        bam_to_sam(
            file2,
            cmp_file2,
            headers=None,
            sorting=Sorting.NAME
        )
        assert_text_files_equal(
            cmp_file1,
            cmp_file2,
            allowed_diff_lines,
            diff_fn=partial(diff_bam_columns, columns=INVARIATE_COLUMNS)
        )

        # Compare subset of reads using all columns
        cmp_file1 = temp / "subset_reads_all_columns_file1"
        cmp_file2 = temp / "subset_reads_all_columns_file2"
        bam_to_sam(
            file1,
            cmp_file1,
            min_mapq=min_mapq,
            sorting=Sorting.COORDINATE,
        )
        bam_to_sam(
            file2,
            cmp_file2,
            min_mapq=min_mapq,
            sorting=Sorting.COORDINATE
        )
        if compare_tag_columns:
            diff_fn = diff_default
        else:
            diff_fn = partial(diff_bam_columns, columns=ALL_COLUMNS)
        assert_text_files_equal(
            cmp_file1,
            cmp_file2,
            allowed_diff_lines,
            diff_fn=diff_fn
        )


def bam_to_sam(
    input_bam: Path,
    output_sam: Path,
    headers: Optional[Iterable[str]] = ("HD", "SQ", "RG"),
    min_mapq: Optional[int] = None,
    sorting: Sorting = Sorting.NONE
):
    """
    Use PySAM to convert bam to sam.
    """
    opts = []
    if headers:
        opts.append("-h")
        headers = set(headers)
    if min_mapq:
        opts.extend(["-q", str(min_mapq)])
    sam = pysam.view(*opts, str(input_bam)).rstrip()
    # Replace any randomly assigned readgroups with a common placeholder
    sam = re.sub(r"UNSET-\w*\b", "UNSET-placeholder", sam)

    lines = sam.splitlines(keepends=True)
    header_lines = []
    start = 0
    if headers:
        for i, line in enumerate(lines):
            if not line.startswith("@"):
                start = i
                break
            elif line[1:3] in headers:
                header_lines.append(line)

    body_lines = lines[start:]
    if sorting is not Sorting.NONE:
        with tempdir() as temp:
            temp_sam = temp / f"output_{str(output_sam.stem)}.sam"
            with open(temp_sam, "w") as out:
                out.write("".join(body_lines))
            if sorting is Sorting.COORDINATE:
                sort_cols = "-k3,3 -k4,4n -k2,2n"
            else:
                sort_cols = "-k1,1 -k2,2n"
            sorted_sam = subby.sub(f"cat {str(temp_sam)} | sort {sort_cols}")
            body_lines = [sorted_sam]

    with open(output_sam, "w") as out:
        out.write("".join(header_lines + body_lines))


def diff_bam_columns(file1: Path, file2: Path, columns: str) -> int:
    with tempdir() as temp:
        def make_comparable(inpath, output):
            subby.run(f"cat {inpath} | cut -f {columns}", stdout=output)

        cmp_file1 = temp / "cmp_file1"
        cmp_file2 = temp / "cmp_file2"
        make_comparable(file1, cmp_file1)
        make_comparable(file2, cmp_file2)
        return diff_default(cmp_file1, cmp_file2)



import logging
from typing import Sequence, Optional
import urllib.request
from urllib.request import Request

from pytest_wdl.url_schemes import Response, BaseResponse, UrlHandler, Method
from pytest_wdl.localizers import download_file
from pytest_wdl.utils import tempdir

LOG = logging.getLogger(__name__)


class MockResponse(BaseResponse):
    def __init__(self, url):
        self.content = url[6:]
        self.start = 0

    def get_content_length(self) -> Optional[int]:
        return len(self.content)

    def read(self, block_size: int):
        start = self.start
        end = min(start + block_size, len(self.content))
        if start == end:
            return None
        block = self.content[start:end]
        self.start = end
        return block.encode()


class MockHandler(UrlHandler):
    def __init__(self):
        self.request_called = False
        self.response_called = False

    @property
    def scheme(self) -> str:
        return "foo"

    @property
    def handles(self) -> Sequence[Method]:
        return [Method.OPEN, Method.REQUEST, Method.RESPONSE]

    def request(self, request: Request) -> Request:
        self.request_called = True
        return request

    def urlopen(self, request: Request) -> Response:
        return MockResponse(request.get_full_url())

    def response(self, request: Request, response: Response) -> Response:
        self.response_called = True
        return response


def test_url_schemes():
    opener = urllib.request._opener
    handler = MockHandler()
    handler.alias()
    try:
        urllib.request.install_opener(urllib.request.build_opener(handler))

        with tempdir() as d:
            outfile = d / "foo"
            download_file("foo://bar", outfile)
            with open(outfile, "rt") as inp:
                assert inp.read() == "bar"

        assert handler.request_called is True
        assert handler.response_called is True
    finally:
        urllib.request._opener = opener



import json
from pathlib import Path
from typing import IO, Optional, Sequence, Union

import requests
from requests.auth import HTTPBasicAuth

from pytest_wdl.executors import (
    Executor,
    ExecutionFailedError,
    get_target_name,
    read_write_inputs,
)
from pytest_wdl.executors._cromwell import CromwellHelperMixin
from pytest_wdl.utils import LOG, PollingException, poll


DEFAULT_API_URL = "http://localhost:8000/api/workflows/v1"
DEFAULT_POLLING_STEP = 5  # seconds
DEFAULT_POLLING_TIMEOUT = 3600  # seconds
TERMINAL_STATES = ["Succeeded", "Aborted", "Failed"]


class CromwellServerExecutor(Executor, CromwellHelperMixin):
    """
    Manages the running of WDL workflows using a remote Cromwell running in Server mode.

    Args:
        import_dirs: Relative or absolute paths to directories containing WDL
            scripts that should be available as imports.
        cromwell_api_url: The full URL where this cromwell exists
            `http://localhost:8000/api/workflows/v1`
        cromwell_api_username: The username to pass to the cromwell API if protected by
            basic auth
        cromwell_api_password: The password to pass to the cromwell API if protected by
            basic auth
        cromwell_configuration: A config file that will be passed to Cromwell
    """

    def __init__(
        self,
        import_dirs: Optional[Sequence[Path]] = None,
        cromwell_api_url: Optional[str] = DEFAULT_API_URL,
        cromwell_api_username: Optional[str] = None,
        cromwell_api_password: Optional[str] = None,
        cromwell_configuration: Optional[Union[str, Path, dict]] = None,
    ):
        self._import_dirs = import_dirs
        self._cromwell_api_url = cromwell_api_url
        self._cromwell_config_file = cromwell_configuration

        if cromwell_api_username and cromwell_api_password:
            self._auth = HTTPBasicAuth(cromwell_api_username, cromwell_api_password)
        else:
            self._auth = None

    def run_workflow(
        self,
        wdl_path: Path,
        inputs: Optional[dict] = None,
        expected: Optional[dict] = None,
        **kwargs,
    ) -> dict:
        """
        Run a WDL workflow on given inputs, and check that the output matches
        given expected values.

        Args:
            wdl_path: The WDL script to execute.
            inputs: Object that will be serialized to JSON and provided to Cromwell
                as the workflow inputs.
            expected: Dict mapping output parameter names to expected values.
            kwargs: Additional keyword arguments, mostly for debugging:
                * workflow_name: The name of the workflow in the WDL script. If None,
                    the name of the WDL script is used (without the .wdl extension).
                * inputs_file: Path to the Cromwell inputs file to use. Inputs are
                    written to this file only if it doesn't exist.
                * imports_file: Path to the WDL imports file to use. Imports are
                    written to this file only if it doesn't exist.
                * java_args: Additional arguments to pass to Java runtime.
                * cromwell_args: Additional arguments to pass to `cromwell run`.

        Returns:
            Dict of outputs.

        Raises:
            ExecutionFailedError: if there was an error executing Cromwell
            AssertionError: if the actual outputs don't match the expected outputs
        """
        target, is_task = get_target_name(
            wdl_path=wdl_path, import_dirs=self._import_dirs, **kwargs
        )

        if is_task:
            raise ValueError(
                "Cromwell cannot execute tasks independently of a workflow"
            )

        inputs_dict, _ = read_write_inputs(
            inputs_file=kwargs.get("inputs_file"),
            inputs_dict=inputs,
            namespace=target,
            write_formatted_inputs=False
        )

        payload = {}
        payload_files = []

        def open_payload_file(path: Path, mode: str = "r") -> IO:
            open_file = open(path, mode)
            payload_files.append(open_file)
            return open_file

        try:
            payload["workflowSource"] = open_payload_file(wdl_path)

            if inputs_dict:
                payload["workflowInputs"] = json.dumps(inputs_dict, default=str)

            imports_file = self._get_workflow_imports(
                self._import_dirs, kwargs.get("imports_file")
            )

            if imports_file:
                payload["workflowDependencies"] = open_payload_file(imports_file, "rb")

            if self._cromwell_config_file:
                if isinstance(inputs_dict, dict):
                    payload["workflowOptions"] = json.dumps(
                        self._cromwell_config_file, default=str
                    )
                else:
                    payload["workflowOptions"] = open_payload_file(
                        self._cromwell_config_file
                    )

            LOG.info(
                f"Executing cromwell server '{self._cromwell_api_url}' with inputs "
                f"{json.dumps(inputs_dict, default=str)}"
            )

            with requests.post(
                self._cromwell_api_url, files=payload, auth=self._auth
            ) as resp:
                status_object = self._resp_to_json(resp, target, inputs_dict)
                run_id = status_object["id"]
                LOG.info(
                    f"Executing on cromwell with id {run_id}. Waiting until terminal "
                    f"state is reached"
                )
        finally:
            for fh in payload_files:
                try:
                    fh.close()
                except:
                    LOG.exception("Error closing file %s", fh)

        self._poll_until_terminal(
            run_id, target, inputs_dict, kwargs.get("timeout", DEFAULT_POLLING_TIMEOUT)
        )

        metadata_url = f"{self._cromwell_api_url}/{run_id}/metadata"
        outputs = None

        with requests.get(metadata_url, auth=self._auth) as metadata_response:
            metadata = self._resp_to_json(metadata_response, target, inputs_dict)

            if metadata["status"] == "Succeeded":
                outputs = metadata["outputs"]
            else:
                error_kwargs = {
                    "executor": "cromwell",
                    "target": target,
                    "status": "Failed",
                    "inputs": inputs_dict,
                }
                self._parse_metadata_errors(
                    metadata, target=target, error_kwargs=error_kwargs
                )
                raise ExecutionFailedError(**error_kwargs)

        if expected:
            self._validate_outputs(outputs, expected, target)

        return outputs

    @staticmethod
    def _resp_to_json(resp, target=None, inputs_dict=None):
        if resp.ok:
            return resp.json()
        else:
            error_kwargs = {
                "executor": "cromwell-server",
                "target": target,
                "status": "Failed",
                "inputs": inputs_dict,
            }

            if resp.reason:
                error_kwargs["msg"] = resp.reason

            raise ExecutionFailedError(**error_kwargs)

    def _poll_until_terminal(
        self,
        run_id: str,
        target: str,
        inputs_dict: Optional[dict] = None,
        timeout: int = DEFAULT_POLLING_TIMEOUT
    ):
        def get_status(status_url):
            with requests.get(status_url, auth=self._auth) as rsp:
                status_dict = self._resp_to_json(rsp, target, inputs_dict)
                return status_dict.get("status") in TERMINAL_STATES

        try:
            poll(
                get_status,
                args=(f"{self._cromwell_api_url}/{run_id}/status",),
                step=DEFAULT_POLLING_STEP,
                timeout=timeout
            )
        except PollingException:
            LOG.exception(f"Encountered timeout for run with id {run_id}")

            error_kwargs = {
                "executor": "cromwell-server",
                "target": target,
                "status": "Failed",
                "inputs": inputs_dict,
                "msg": f"Encountered timeout for run with id {run_id}",
            }

            raise ExecutionFailedError(**error_kwargs)



import json
import re

from pytest_wdl.config import UserConfiguration
from pytest_wdl.utils import tempdir
from . import setenv


def test_user_config_no_defaults():
    with tempdir(True):
        config = UserConfiguration()
        assert config.remove_cache_dir is True
        assert config.cache_dir.exists()
        config.cleanup()
        assert not config.cache_dir.exists()


def test_user_config_from_file():
    with tempdir() as d, setenv({
        "HTTPS_PROXY": "http://foo.com/https",
        "FOO_HEADER": "bar"
    }):
        cache_dir = d / "cache"
        execution_dir = d / "execution"
        config_dict = {
            "cache_dir": str(cache_dir),
            "execution_dir": str(execution_dir),
            "proxies": {
                "http": {
                    "value": "http://foo.com/http"
                },
                "https": {
                    "env": "HTTPS_PROXY"
                }
            },
            "http_headers": [
                {
                    "pattern": "http://foo.com/.*",
                    "name": "foo",
                    "env": "FOO_HEADER"
                }
            ],
            "executors": {
                "foo": {
                    "bar": 1
                }
            }
        }
        config_file = d / "config.json"
        with open(config_file, "wt") as out:
            json.dump(config_dict, out)
        config = UserConfiguration(config_file)
        assert config.cache_dir == cache_dir
        assert config.default_execution_dir == execution_dir
        assert config.proxies == {
            "http": "http://foo.com/http",
            "https": "http://foo.com/https"
        }
        assert config.default_http_headers == [
            {
                "pattern": re.compile("http://foo.com/.*"),
                "name": "foo",
                "env": "FOO_HEADER"
            }
        ]
        assert config.get_executor_defaults("foo") == {"bar": 1}



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json
import re
import pytest
from pytest_wdl.config import UserConfiguration
from pytest_wdl.localizers import (
    LinkLocalizer, StringLocalizer, JsonLocalizer, UrlLocalizer
)
from pytest_wdl.utils import DigestsNotEqualError, tempdir
from . import GOOD_URL, no_internet, setenv


def test_link_localizer():
    with tempdir() as d:
        foo = d / "foo"
        with open(foo, "wt") as out:
            out.write("foo")
        bar = d / "bar"
        localizer = LinkLocalizer(foo)
        localizer.localize(bar)
        assert bar.exists()
        assert bar.is_symlink()


def test_string_localizer():
    with tempdir() as d:
        foo = d / "foo"
        StringLocalizer("foo").localize(foo)
        with open(foo, "rt") as inp:
            assert inp.read() == "foo"


def test_json_localizer():
    with tempdir() as d:
        foo = d / "foo"
        contents = {
            "foo": 1,
            "bar": "a"
        }
        JsonLocalizer(contents).localize(foo)
        with open(foo, "rt") as inp:
            assert json.load(inp) == contents


@pytest.mark.skipif(no_internet, reason="no internet available")
def test_url_localizer():
    good_url = GOOD_URL
    bad_url = "foo"
    with tempdir() as d:
        foo = d / "foo"
        UrlLocalizer(
            good_url,
            UserConfiguration(None, cache_dir=d),
            digests={"md5": "acbd18db4cc2f85cedef654fccc4a4d8"}
        ).localize(foo)
        with open(foo, "rt") as inp:
            assert inp.read() == "foo"

        # test that the file is overwritten
        with open(foo, "wt") as out:
            out.write("bork")
        UrlLocalizer(good_url, UserConfiguration(None, cache_dir=d)).localize(foo)
        with open(foo, "rt") as inp:
            assert inp.read() == "foo"

    with pytest.raises(RuntimeError):
        with tempdir() as d:
            foo = d / "foo"
            UrlLocalizer(
                good_url,
                UserConfiguration(None, cache_dir=d),
                digests={"md5": "XXX"}
            ).localize(foo)

    with pytest.raises(RuntimeError):
        UrlLocalizer(bad_url, UserConfiguration(None, cache_dir=d)).localize(foo)


# @pytest.mark.skipif(no_internet, reason="no internet available")
# def test_url_localizer_corrupt_file():
#     good_url = GOOD_URL
#     with tempdir() as d:
#         foo = d / "foo"
#         # The localizer should detect that the file already exists, but that the
#         # digests don't match and overwrite it
#         with open(foo, "wt") as out:
#             out.write("blahblahblah")
#         UrlLocalizer(
#             good_url,
#             UserConfiguration(None, cache_dir=d),
#             digests={"md5": "acbd18db4cc2f85cedef654fccc4a4d8"}
#         ).localize(foo)
#         with open(foo, "rt") as inp:
#             assert inp.read() == "foo"


def test_url_localizer_add_headers():
    with setenv({
       "FOO": "bar"
    }):
        localizer = UrlLocalizer(
            "http://foo.com/bork",
            UserConfiguration(http_headers=[
                {
                    "name": "beep",
                    "pattern": re.compile(r"http://foo.com/.*"),
                    "env": "FOO",
                    "value": "baz"
                },
                {
                    "name": "boop",
                    "pattern": re.compile(r"http://foo.*/bork"),
                    "env": "BAR",
                    "value": "blorf"
                }
            ]),
            {
                "boop": {
                    "value": "blammo"
                }
            }
        )
        headers = localizer.http_headers
        assert len(headers) == 2
        assert set(headers.keys()) == {"beep", "boop"}
        assert headers["beep"] == "bar"
        assert headers["boop"] == "blammo"


def test_url_localizer_set_proxies():
    localizer = UrlLocalizer("http://foo.com", UserConfiguration(proxies={
        "https": "https://foo.com/proxy"
    }))
    proxies = localizer.proxies
    assert len(proxies) == 1
    assert "https" in proxies
    assert proxies["https"] == "https://foo.com/proxy"



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json
from pathlib import Path
from typing import cast

import pytest

from pytest_wdl.core import DefaultDataFile, create_executor
from pytest_wdl.executors import (
    Executor, ExecutionFailedError, InputsFormatter, read_write_inputs
)
from pytest_wdl.utils import tempdir


WORKFLOW_EXECUTORS = ["cromwell", "miniwdl"]
TASK_EXECUTORS = ["miniwdl"]


@pytest.mark.integration
@pytest.mark.parametrize("executor", WORKFLOW_EXECUTORS)
def test_executors(workflow_data, workflow_runner, executor):
    inputs = {
        "in_txt": workflow_data["in_txt"],
        "in_int": 1
    }
    expected = {
        "out_txt": workflow_data["out_txt"],
        "out_int": 1
    }
    workflow_runner(
        "test.wdl",
        inputs,
        expected,
        executors=[executor]
    )
    # Test with the old workflow_runner signature
    workflow_runner(
        "test.wdl",
        "cat_file",
        inputs,
        expected,
        executors=[executor]
    )


@pytest.mark.integration
def test_multiple_executors(workflow_data, workflow_runner):
    inputs = {
        "in_txt": workflow_data["in_txt"],
        "in_int": 1
    }
    expected = {
        "out_txt": workflow_data["out_txt"],
        "out_int": 1
    }

    def callback(executor: str, execution_dir: Path, outputs: dict):
        assert executor in WORKFLOW_EXECUTORS
        assert execution_dir.exists()
        for param in ("cat_file.out_txt", "cat_file.out_int"):
            assert param in outputs

    workflow_runner(
        "test.wdl",
        inputs,
        expected,
        executors=WORKFLOW_EXECUTORS,
        callback=callback
    )


# TODO: figure out how to override default_executors
# def test_workflow_runner_error(workflow_data, workflow_runner):
#     inputs = {
#         "in_txt": workflow_data["in_txt"],
#         "in_int": 1
#     }
#     outputs = {
#         "out_txt": workflow_data["out_txt"],
#         "out_int": 1
#     }
#
#     with pytest.raises(RuntimeError):
#         workflow_runner(
#             "test.wdl",
#             inputs,
#             outputs,
#             executors=[]
#         )


@pytest.mark.integration
@pytest.mark.parametrize("executor", TASK_EXECUTORS)
def test_task(workflow_data, workflow_runner, executor):
    inputs = {
        "in_txt": workflow_data["in_txt"],
    }
    outputs = {
        "out_txt": workflow_data["out_txt"],
    }
    workflow_runner(
        "test.wdl",
        inputs,
        outputs,
        executors=[executor],
        task_name="cat"
    )


@pytest.mark.integration
@pytest.mark.parametrize("executor", WORKFLOW_EXECUTORS)
def test_execution_failure(workflow_data, workflow_runner, executor):
    inputs = {
        "in_txt": workflow_data["in_txt"],
        "in_int": 1,
        "fail": True
    }
    outputs = {
        "out_txt": workflow_data["out_txt"],
        "out_int": 1
    }
    with pytest.raises(ExecutionFailedError) as exc_info:
        workflow_runner(
            "test.wdl",
            inputs,
            outputs,
            executors=[executor]
        )

    err = cast(ExecutionFailedError, exc_info.value)
    assert "foo_fail" in err.failed_task
    assert err.failed_task_exit_status == 1


def test_get_workflow_inputs():
    actual_inputs_dict, inputs_path = read_write_inputs(
        inputs_dict={"bar": 1}, namespace="foo"
    )
    assert inputs_path.exists()
    with open(inputs_path, "rt") as inp:
        assert json.load(inp) == actual_inputs_dict
    assert actual_inputs_dict == {
        "foo.bar": 1
    }

    with tempdir() as d:
        inputs_file = d / "inputs.json"
        actual_inputs_dict, inputs_path = read_write_inputs(
            inputs_dict={"bar": 1}, namespace="foo", inputs_file=inputs_file
        )
        assert inputs_file == inputs_path
        assert inputs_path.exists()
        with open(inputs_path, "rt") as inp:
            assert json.load(inp) == actual_inputs_dict
        assert actual_inputs_dict == {
            "foo.bar": 1
        }

    with tempdir() as d:
        inputs_file = d / "inputs.json"
        inputs_dict = {"foo.bar": 1}
        with open(inputs_file, "wt") as out:
            json.dump(inputs_dict, out)
        actual_inputs_dict, inputs_path = read_write_inputs(
            namespace="foo", inputs_file=inputs_file
        )
        assert inputs_file == inputs_path
        assert inputs_path.exists()
        with open(inputs_path, "rt") as inp:
            assert json.load(inp) == actual_inputs_dict
        assert actual_inputs_dict == inputs_dict


def test_inputs_formatter():
    formatter = InputsFormatter.get_instance()

    assert formatter.format_value(1) == 1
    assert formatter.format_value("foo") == "foo"
    assert formatter.format_value((1.1, 2.2)) == [1.1, 2.2]

    with tempdir() as d:
        foo = d / "foo"
        with open(foo, "wt") as out:
            out.write("foo")
        df = DefaultDataFile(foo)
        assert formatter.format_value(df) == foo
        assert formatter.format_value([df]) == [foo]
        assert formatter.format_value({"a": df}) == {"a": foo}

    class Obj:
        def __init__(self, a: str, b: int):
            self.a = a
            self.b = b

        def as_dict(self):
            return {
                "a": self.a,
                "b": self.b
            }

    assert formatter.format_value(Obj("hi", 1)) == {"a": "hi", "b": 1}


def test_create_executor():
    with pytest.raises(RuntimeError):
        create_executor("foo", [], None)


def test_validate_outputs():
    with pytest.raises(AssertionError):
        Executor._validate_outputs(
            {"foo.bar": 1}, {"bar": 1}, ""
        )
    with pytest.raises(AssertionError):
        Executor._validate_outputs(
            {"foo.bar": 1}, {"baz": 1}, "foo"
        )
    Executor._validate_outputs(
        {"foo.bar": None}, {"bar": None}, "foo"
    )
    with pytest.raises(AssertionError):
        Executor._validate_outputs(
            {"foo.bar": None}, {"bar": 1}, "foo"
        )
    with pytest.raises(AssertionError):
        Executor._validate_outputs(
            {"foo.bar": 1}, {"bar": None}, "foo"
        )
    Executor._validate_outputs(
        {"foo.bar": [1, 2, 3]}, {"bar": [1, 2, 3]}, "foo"
    )
    with pytest.raises(AssertionError):
        Executor._validate_outputs(
            {"foo.bar": [1, 2, 3]}, {"bar": [1, 2]}, "foo"
        )
    with pytest.raises(AssertionError):
        Executor._validate_outputs(
            {"foo.bar": [1, 2, 3]}, {"bar": [3, 2, 1]}, "foo"
        )
    Executor._validate_outputs(
        {"foo.bar": {"a": 1}}, {"bar": {"a": 1}}, "foo"
    )
    with pytest.raises(AssertionError):
        Executor._validate_outputs(
            {"foo.bar": {"a": 1}}, {"bar": {"b": 1}}, "foo"
        )
    with pytest.raises(AssertionError):
        Executor._validate_outputs(
            {"foo.bar": {"a": 1}}, {"bar": {"a": 1, "b": 2}}, "foo"
        )
    with pytest.raises(AssertionError):
        Executor._validate_outputs(
            {"foo.bar": {"a": 1}}, {"bar": {"a": 2}}, "foo"
        )
    with pytest.raises(AssertionError):
        Executor._validate_outputs(
            {"foo.bar": 1}, {"bar": 2}, "foo"
        )



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import os
from pathlib import Path
import tempfile
from typing import Any, Callable, List, Optional, Sequence, Type, Union, cast

from pytest_wdl.config import UserConfiguration
from pytest_wdl.data_types import DEFAULT_TYPE, DataFile, DefaultDataFile
from pytest_wdl.executors import Executor
from pytest_wdl.localizers import (
    LinkLocalizer, StringLocalizer, JsonLocalizer, UrlLocalizer
)
from pytest_wdl.plugins import plugin_factory_map
from pytest_wdl.url_schemes import install_schemes
from pytest_wdl.utils import ensure_path


DATA_TYPES = plugin_factory_map(DataFile, "pytest_wdl.data_types")
"""Data type plugin modules from the discovered entry points."""

EXECUTORS = plugin_factory_map(Executor, "pytest_wdl.executors")
"""Executor plugin modules from the discovered entry points."""

# Install URL scheme plugins
install_schemes()


class DataDirs:
    """
    Provides data files from test data directory structure as defined by the
    datadir and datadir-ng plugins. Paths are resolved lazily upon first request.
    """
    def __init__(
        self,
        basedir: Path,
        module: Optional[Union[str, Any]] = None,  # TODO: no Module type in typelib yet
        function: Optional[Union[str, Callable]] = None,
        cls: Optional[Union[str, Type]] = None
    ):
        self.basedir = basedir

        # If there are packages in the tests/ folder (i.e. if there are __init__.py
        # files), we need to drop any packages from the module_path since they
        # will conflict with basedir.
        if module:
            if isinstance(module, str):
                module_name = cast(str, module)
            else:
                module_name = module.__name__

            module_path = module_name.split(".")
            num_pkgs = len(module_path) - 1

            if num_pkgs > 0:
                basedir_parts = basedir.parts

                if (
                    len(basedir_parts) < num_pkgs
                    or tuple(basedir_parts[-num_pkgs:]) != tuple(module_path[:num_pkgs])
                ):
                    raise RuntimeError(
                        f"Module path {module_path} does not match basedir {basedir}"
                    )

            self.module = self.module = module_path[-1]
        else:
            self.module = None

        if not function:
            self.function = None
        elif isinstance(function, str):
            self.function = function
        else:
            self.function = cast(Callable, function).__name__

        if not cls:
            self.cls = None
        elif isinstance(cls, str):
            self.cls = cast(str, cls)
        else:
            self.cls = cast(Type, cls).__name__

        self._paths = None

    @property
    def paths(self) -> List[Path]:
        if self._paths is None:
            def add_datadir_paths(root: Path):
                if root.exists():
                    if self.cls:
                        clsdir = root / self.cls

                        if clsdir.exists():
                            if self.function:
                                fndir = clsdir / self.function

                                if fndir.exists():
                                    self._paths.append(fndir)

                            self._paths.append(clsdir)
                    elif self.function:
                        fndir = root / self.function

                        if fndir.exists():
                            self._paths.append(fndir)

                    self._paths.append(root)

            self._paths = []

            if self.module:
                add_datadir_paths(self.basedir / self.module)
                add_datadir_paths(self.basedir / "data" / self.module)

            add_datadir_paths(self.basedir / "data")

        return self._paths


class DataResolver:
    """
    Resolves data files that may need to be localized.
    """
    def __init__(self, data_descriptors: dict, user_config: UserConfiguration):
        self.data_descriptors = data_descriptors
        self.user_config = user_config

    def resolve(self, name: str, datadirs: Optional[DataDirs] = None):
        if name in self.data_descriptors:
            value = self.data_descriptors[name]

            if isinstance(value, dict):
                # Right now, "class" is just a marker for object types, of which
                # "file" is a special case.
                cls = value.get("class", "file")
                if "value" in value:
                    value = value["value"]
                if cls == "file":
                    value = create_data_file(
                        user_config=self.user_config,
                        datadirs=datadirs,
                        **cast(dict, value)
                    )
        else:
            value = create_data_file(
                name=name,
                user_config=self.user_config,
                datadirs=datadirs
            )

        return value


class DataManager:
    """
    Manages test data, which is defined in a test_data.json file.

    Args:
        data_resolver: Module-level config.
        datadirs: Data directories to search for the data file.
    """
    def __init__(self, data_resolver: DataResolver, datadirs: DataDirs):
        self.data_resolver = data_resolver
        self.datadirs = datadirs

    def __getitem__(self, name: str):
        return self.data_resolver.resolve(name, self.datadirs)

    def get_list(self, *names: str) -> list:
        return [self[name] for name in names]

    def get_dict(self, *names: str, **params) -> dict:
        """
        Creates a dict with one or more entries from this DataManager.

        Args:
            *names: Names of test data entries to add to the dict.
            **params: Mapping of workflow parameter names to test data entry names.

        Returns:
            Dict mapping parameter names to test data entries for all specified names.
        """
        d = {}
        for name in names:
            d[name] = self[name]
        for param, name in params.items():
            d[param] = self[name]
        return d


def create_data_file(
    user_config: UserConfiguration,
    type: Optional[Union[str, dict]] = DEFAULT_TYPE,
    name: Optional[str] = None,
    path: Optional[Union[str, Path]] = None,
    url: Optional[str] = None,
    contents: Optional[Union[str, dict]] = None,
    env: Optional[str] = None,
    http_headers: Optional[dict] = None,
    digests: Optional[dict] = None,
    datadirs: Optional[DataDirs] = None,
    **kwargs
) -> DataFile:
    if isinstance(type, dict):
        data_file_opts = cast(dict, type)
        type = data_file_opts.pop("name")
    else:
        data_file_opts = {}

    data_file_opts.update(kwargs)

    local_path = None
    localizer = None

    if path:
        local_path = ensure_path(path, [user_config.cache_dir])

    if local_path and local_path.exists():
        pass
    elif env and env in os.environ:
        env_path = ensure_path(os.environ[env], exists=True)
        if not local_path:
            local_path = env_path
        else:
            localizer = LinkLocalizer(env_path)
    elif url:
        localizer = UrlLocalizer(url, user_config, http_headers, digests)
        if not local_path:
            if name:
                local_path = ensure_path(user_config.cache_dir / name)
            else:
                filename = url.rsplit("/", 1)[1]
                local_path = ensure_path(user_config.cache_dir / filename)
    elif contents:
        if isinstance(contents, str):
            localizer = StringLocalizer(cast(str, contents))
        else:
            localizer = JsonLocalizer(cast(dict, contents))
            if type == DEFAULT_TYPE:
                type = "json"
        if not local_path:
            if name:
                local_path = ensure_path(user_config.cache_dir / name)
            else:
                local_path = ensure_path(
                    tempfile.mktemp(dir=user_config.cache_dir)
                )
    elif name and datadirs:
        for dd in datadirs.paths:
            dd_path = dd / name
            if dd_path.exists():
                break
        else:
            raise FileNotFoundError(
                f"File {name} not found in any of the following datadirs: "
                f"{datadirs.paths}"
            )

        if not local_path:
            local_path = dd_path
        else:
            localizer = LinkLocalizer(dd_path)
    else:
        raise FileNotFoundError(
            f"File {path or name} does not exist. Either a url, file contents, "
            f"or a local file must be provided."
        )

    data_file_class = DATA_TYPES.get(type, DefaultDataFile)
    return data_file_class(local_path, localizer, **data_file_opts)


def create_executor(
    executor_name: str,
    import_dirs: Sequence[Path],
    user_config: UserConfiguration
):
    executor_class = EXECUTORS.get(executor_name)
    if not executor_class:
        raise RuntimeError(f"{executor_name} executor plugin is not installed")
    return executor_class(
        import_dirs,
        **user_config.get_executor_defaults(executor_name)
    )



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Callable, Optional, Union, cast

import subby

from pytest_wdl.localizers import Localizer
from pytest_wdl.utils import compare_files_with_hash, ensure_path, tempdir
from xphyle import guess_file_format
from xphyle.utils import transcode_file

DEFAULT_TYPE = "default"
ALLOWED_DIFF_LINES = "allowed_diff_lines"


class DataFile(metaclass=ABCMeta):
    """
    A data file, which may be local, remote, or represented as a string.

    Args:
        local_path: Path where the data file should exist after being localized.
        localizer: Localizer object, for persisting the file on the local disk.
        allowed_diff_lines: Number of lines by which the file is allowed to differ
            from another and still be considered equal.
        compare_opts: Additional type-specific comparison options.
    """
    def __init__(
        self,
        local_path: Path,
        localizer: Optional[Localizer] = None,
        **compare_opts
    ):
        if localizer is None and not local_path.exists():
            raise ValueError(
                f"Local path {local_path} does not exist and 'localizer' is None"
            )
        self.local_path = local_path
        self.localizer = localizer
        self.compare_opts = compare_opts

    @property
    def path(self) -> Path:
        if not self.local_path.exists():
            if self.localizer:
                ensure_path(self.local_path, is_file=True, create=True)
                self.localizer.localize(self.local_path)
            else:
                raise RuntimeError(
                    f"Localization to {self.local_path} is required but no localizer "
                    f"is defined"
                )
        return self.local_path

    def __str__(self) -> str:
        return str(self.local_path)

    def set_compare_opts(self, **kwargs):
        """
        Update comparison options.

        Args:
            **kwargs: Comparison options to update.
        """
        self.compare_opts.update(kwargs)

    def assert_contents_equal(self, other: Union[str, Path, "DataFile"]) -> None:
        """
        Assert the contents of two files are equal.

        If `allowed_diff_lines == 0`, files are compared using MD5 hashes, otherwise
        their contents are compared using the linux `diff` command.

        Args:
            other: A `DataFile` or string file path.

        Raises:
            AssertionError if the files are different.
        """
        other_compare_opts = {}
        if isinstance(other, Path):
            other_path = other
        elif isinstance(other, str):
            other_path = Path(other)
        else:
            other_df = cast(DataFile, other)
            other_path = other_df.path
            other_compare_opts = other.compare_opts

        self._assert_contents_equal(other_path, other_compare_opts)

    @abstractmethod
    def _assert_contents_equal(self, other_path: Path, other_opts: dict) -> None:
        pass

    def _get_allowed_diff_lines(self, other_opts: dict):
        return max(
            self.compare_opts.get(ALLOWED_DIFF_LINES, 0),
            other_opts.get(ALLOWED_DIFF_LINES, 0)
        )


class DefaultDataFile(DataFile):
    def _assert_contents_equal(self, other_path: Path, other_opts: dict):
        allowed_diff_lines = self._get_allowed_diff_lines(other_opts)
        if allowed_diff_lines:
            assert_text_files_equal(self.path, other_path, allowed_diff_lines)
        else:
            assert_binary_files_equal(self.path, other_path)


def diff_default(file1: Path, file2: Path) -> int:
    """
    Default diff command.

    Args:
        file1: First file to compare
        file2: Second file to compare

    Returns:
        Number of different lines.
    """
    with tempdir() as temp:
        # Remove trailing whitespace, and ensure a newline at the end of the file
        cmp_file1 = temp / "file1"
        cmp_file2 = temp / "file2"
        subby.run("sed 's/[[:space:]]*$//; $a\\'", stdin=file1, stdout=cmp_file1)
        subby.run("sed 's/[[:space:]]*$//; $a\\'", stdin=file2, stdout=cmp_file2)

        # diff - it would be possible to do this without sed using GNU diff with the
        # `--ignore-trailing-space` option, but unfortunately that option is not
        # available in macOS diff, which provides BSD versions of the tools by default.
        cmds = [
            f"diff -y --suppress-common-lines {cmp_file1} {cmp_file2}",
            "grep -c '^'"
        ]

        # It's a valid result to have no lines match, so allow a grep returncode of 1
        return int(subby.sub(cmds, allowed_return_codes=(0, 1)))


def assert_text_files_equal(
    file1: Path,
    file2: Path,
    allowed_diff_lines: int = 0,
    diff_fn: Callable[[Path, Path], int] = diff_default
) -> None:
    fmt = guess_file_format(file1)
    if fmt:
        with tempdir() as temp:
            temp_file1 = temp / "file1"
            temp_file2 = temp / "file2"
            transcode_file(file1, temp_file1, dest_compression=False)
            transcode_file(file2, temp_file2, dest_compression=False)
            diff_lines = diff_fn(temp_file1, temp_file2)
    else:
        diff_lines = diff_fn(file1, file2)

    if diff_lines > allowed_diff_lines:
        raise AssertionError(
            f"{diff_lines} lines (which is > {allowed_diff_lines} allowed) are "
            f"different between files {file1}, {file2}"
        )


def compare_gzip(file1: Path, file2: Path):
    crc_size1 = subby.sub(f"gzip -lv {file1} | tail -1 | awk '{{print $2\":\"$7}}'")
    crc_size2 = subby.sub(f"gzip -lv {file2} | tail -1 | awk '{{print $2\":\"$7}}'")
    if crc_size1 != crc_size2:  # TODO: test this
        raise AssertionError(
            f"CRCs and/or uncompressed sizes differ between expected identical "
            f"gzip files {file1}, {file2}"
        )


# TODO: allow user-defined comparators
BINARY_COMPARATORS = {
    "gz": compare_gzip,
    "gzip": compare_gzip
}


def assert_binary_files_equal(file1: Path, file2: Path, digest: str = "md5") -> None:
    fmt = guess_file_format(file1)
    if fmt and fmt in BINARY_COMPARATORS:
        BINARY_COMPARATORS[fmt](file1, file2)
    else:
        compare_files_with_hash(file1, file2, digest)



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Fixtures for writing tests that execute WDL workflows using Cromwell.

Note: This library is being transitioned to python3 only, and to use `pathlib.Path`s
instead of string paths. For backward compatibility fixtures that produce a path may
still return string paths, but this support will be dropped in a future version.
"""
import json
from pathlib import Path
from typing import Callable, List, Optional, Sequence, Tuple, Union

from _pytest.fixtures import FixtureRequest
from pytest_subtests import SubTests

from pytest_wdl import config
from pytest_wdl.config import UserConfiguration
from pytest_wdl.core import DataResolver, DataManager, DataDirs, create_executor
from pytest_wdl.utils import ensure_path, context_dir, find_project_path

try:
    import yaml
except ImportError:
    yaml = None


DEFAULT_TEST_DATA_FILE = "test_data"
DEFAULT_IMPORT_PATHS_FILE = "import_paths.txt"


def user_config_file() -> Optional[Path]:
    """
    Fixture that provides the value of 'user_config' environment variable. If
    not specified, looks in the default location ($HOME/pytest_user_config.json).

    Returns:
        Path to the confif file, or None if not specified.
    """
    return config.default_user_config_file()


def user_config(user_config_file: Optional[Path]) -> UserConfiguration:
    config.set_instance(path=user_config_file)
    yield config.get_instance()
    config.cleanup()


def project_root_files() -> List[str]:
    """
    Fixture that provides a list of filenames that are found in the project root
    directory. Used by the `project_root` fixture to locate the project root
    directory.
    """
    return [".git", "setup.py", "pyproject.toml"]


def project_root(
    request: FixtureRequest, project_root_files: List[str]
) -> Union[str, Path]:
    """
    Fixture that provides the root directory of the project. By default, this
    assumes that the project has one subdirectory per task, and that this
    framework is being run from the test subdirectory of a task diretory, and
    therefore looks for the project root two directories up.
    """
    path = Path(request.fspath.dirpath())
    root = find_project_path(*project_root_files, start=path, return_parent=True)
    if root:
        return root
    else:  # TODO: test this
        # If the project root can't be identified, fall back to the parent of
        # the current directory (which is probably tests).
        return path.parent


def workflow_data_descriptor_file(request: FixtureRequest) -> Union[str, Path]:
    """
    Fixture that provides the path to the JSON file that describes test data files.

    Args:
        request: A FixtureRequest object
    """
    test_data_files = [f"{DEFAULT_TEST_DATA_FILE}.json"]

    if yaml:
        test_data_files.append(f"{DEFAULT_TEST_DATA_FILE}.yaml")

    test_data_paths = []

    for f in test_data_files:
        test_data_paths.extend((Path(f), Path("tests") / f))

    return find_project_path(
        *test_data_files, start=Path(request.fspath.dirpath()), assert_exists=True
    )


def workflow_data_descriptors(
    request: FixtureRequest,
    project_root: Union[str, Path],
    workflow_data_descriptor_file: Union[str, Path],
) -> dict:
    """
    Fixture that provides a mapping of test data names to values. If
    workflow_data_descriptor_file is relative, it is searched first relative to the
    current test context directory and then relative to the project root.

    Args:
        workflow_data_descriptor_file: Path to the data descriptor JSON file.

    Returns:
        A dict with keys as test data names and each value either a
        primitive, a map describing a data file, or a DataFile object.
    """
    search_paths = [Path(request.fspath.dirpath()), project_root]
    workflow_data_descriptor_path = ensure_path(
        workflow_data_descriptor_file,
        search_paths=search_paths,
        is_file=True,
        exists=True,
    )
    with open(workflow_data_descriptor_path, "rt") as inp:
        if yaml and workflow_data_descriptor_path.suffix == ".yaml":
            return yaml.load(inp)
        else:
            return json.load(inp)


def workflow_data_resolver(
    workflow_data_descriptors: dict, user_config: UserConfiguration
) -> DataResolver:
    """
    Provides access to test data files for tests in a module.

    Args:
        workflow_data_descriptors: workflow_data_descriptors fixture.
        user_config:
    """
    return DataResolver(workflow_data_descriptors, user_config)


def workflow_data(
    request: FixtureRequest, workflow_data_resolver: DataResolver
) -> DataManager:
    """
    Provides an accessor for test data files, which may be local or in a remote
    repository.

    Args:
        request: FixtureRequest object
        workflow_data_resolver: Module-level test data configuration

    Examples:
        def workflow_data_descriptor_file():
            return "tests/test_data.json"

        def test_workflow(workflow_data):
            print(workflow_data["myfile"])
    """
    data_dirs = DataDirs(
        ensure_path(request.fspath.dirpath(), canonicalize=True),
        request.module,
        request.function,
        request.cls,
    )
    return DataManager(workflow_data_resolver, data_dirs)


def import_paths(request: FixtureRequest) -> Union[str, Path, None]:
    """
    Fixture that provides the path to a file that lists directories containing WDL
    scripts to make available as imports. This looks for the file at
    "tests/import_paths.txt" by default, and returns None if that file doesn't exist.
    """
    import_paths_file = Path(request.fspath.dirpath()) / DEFAULT_IMPORT_PATHS_FILE
    if import_paths_file.exists():
        return import_paths_file


def import_dirs(
    request: FixtureRequest,
    project_root: Union[str, Path],
    import_paths: Optional[Union[str, Path]],
) -> List[Union[str, Path]]:
    """
    Fixture that provides a list of directories containing WDL scripts to make
    avaialble as imports. Uses the file provided by `import_paths` fixture if
    it is not None, otherwise returns a list containing the parent directory
    of the test module.

    Args:
        request: A FixtureRequest object
        project_root: Project root directory
        import_paths: File listing paths to imports, one per line
    """
    if import_paths:
        import_paths = ensure_path(import_paths, canonicalize=True)

        if not import_paths.exists():
            raise FileNotFoundError(f"import_paths file {import_paths} does not exist")

        paths = []

        with open(import_paths, "rt") as inp:
            for path_str in inp.read().splitlines(keepends=False):
                path = Path(path_str)
                if not path.is_absolute():
                    path = ensure_path(project_root / path)
                if not path.exists():
                    raise FileNotFoundError(f"Invalid import path: {path}")
                paths.append(path)

        return paths
    else:
        module_dir = find_project_path(
            "tests", start=Path(request.fspath.dirpath()), return_parent=True
        )

        if module_dir:
            return [module_dir]
        else:
            return []


def default_executors(user_config: UserConfiguration) -> Sequence[str]:
    return user_config.executors


def workflow_runner(
    request: FixtureRequest,
    project_root: Union[str, Path],
    import_dirs: List[Union[str, Path]],
    user_config: UserConfiguration,
    default_executors: Sequence[str],
    subtests: SubTests,
):
    """
    Provides a callable that runs a workflow. The callable has the same signature as
    `Executor.run_workflow`, but takes an additional keyword argument `executors`,
    a sequence of strings, which allows overriding the names of the executors to use.

    If multiple executors are specified, the tests are run using the `subtests`
    fixture of the `pytest-subtests` plugin.

    Args:
        request: A FixtureRequest object.
        project_root: Project root directory.
        import_dirs: Directories from which to import WDL scripts.
        user_config: A UserConfiguration object.
        default_executors: Names of executors to use when executor name isn't passed to
            the `workflow_runner` callable.
        subtests: A SubTests object.

    Returns:
        A generator over the results of calling the workflow with each executor. Each
        value is a tuple `(executor_name, execution_dir, outputs)`, where
        `execution_dir` is the root directory where the task/workflow was run (the
        structure of the directory is executor-dependent) and `outputs` is a dict of
        the task/workflow outputs.
    """
    return WorkflowRunner(
        default_executors=default_executors,
        wdl_search_paths=[Path(request.fspath.dirpath()), project_root],
        import_dirs=[ensure_path(d, is_file=False, exists=True) for d in import_dirs],
        user_config=user_config,
        subtests=subtests,
    )


class WorkflowRunner:
    """
    Callable object that runs tests.
    """
    def __init__(
        self,
        wdl_search_paths: Sequence[Path],
        import_dirs: Sequence[Path],
        user_config: UserConfiguration,
        subtests: SubTests,
        default_executors: Optional[Sequence[str]] = None
    ):
        """
        Base class for test runners.
        """
        self._wdl_search_paths = wdl_search_paths
        self._import_dirs = import_dirs
        self._user_config = user_config
        self._subtests = subtests
        self._default_executors = default_executors or ()

    def __call__(self, *args, **kwargs) -> dict:
        executors, call_args = self._args(*args, **kwargs)

        outputs = {}
        failed = []

        if len(executors) == 1:
            outputs[executors[0]] = self._run_test(executors[0], **call_args)
        else:
            for executor_name in executors:
                with self._subtests.test(executor_name=executor_name):
                    try:
                        outputs[executor_name] = self._run_test(
                            executor_name, **call_args
                        )
                    except:
                        failed.append(executor_name)
                        raise

        if failed:
            raise AssertionError(f"One or more sub-tests failed: {failed}")

        return outputs

    def _args(
        self,
        wdl_script: Union[str, Path],
        *args,
        inputs: Optional[dict] = None,
        expected: Optional[dict] = None,
        executors: Optional[Sequence[str]] = None,
        **kwargs,
    ) -> Tuple[Sequence[str], dict]:
        """
        Handle multiple different call signatures.

        Args:
            wdl_script:
            args: Positional arguments. Supports backward-compatibility for workflows
                using the old `run_workflow` signature in which the second argument
                was the workflow name. This will be removed in the next major version.
            inputs:
            expected:
            executors:
            kwargs: Additional keyword arguments

        Returns:
            Tuple of (executors, call_kwargs).
        """
        wdl_path = ensure_path(
            wdl_script, self._wdl_search_paths, is_file=True, exists=True
        )

        if args:
            args_list = list(args)

            if isinstance(args_list[0], str):
                kwargs["workflow_name"] = args_list.pop(0)

            if args_list:
                if inputs:
                    raise TypeError("Multiple values for argument 'inputs'")

                inputs = args_list.pop(0)

                if args_list:
                    if expected:
                        raise TypeError("Multiple values for argument 'expected'")

                    expected = args_list.pop(0)

                    if args_list:
                        raise TypeError("Too many arguments")

        if not executors:
            executors = self._default_executors

        if len(executors) == 0:
            raise RuntimeError("At least one executor must be specified")

        call_args = {
            "wdl_path": wdl_path,
            "inputs": inputs,
            "expected": expected,
        }

        call_args.update(kwargs)

        return executors, call_args

    def _run_test(
        self,
        executor_name: str,
        wdl_path: Path,
        inputs: Optional[dict] = None,
        expected: Optional[dict] = None,
        callback: Optional[Callable[[str, Path, dict], None]] = None,
        **kwargs
    ) -> dict:
        executor = create_executor(executor_name, self._import_dirs, self._user_config)

        with context_dir(
            self._user_config.default_execution_dir, change_dir=True
        ) as execution_dir:
            outputs = executor.run_workflow(
                wdl_path, inputs=inputs, expected=expected, **kwargs
            )

            if callback:
                callback(executor_name, execution_dir, outputs)

            return outputs



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import glob
import json
from pathlib import Path
import re
import tempfile
from typing import Optional, Sequence, Union, cast

import subby

from pytest_wdl.utils import LOG, ensure_path

ENV_CROMWELL_JAR = "CROMWELL_JAR"
ENV_CROMWELL_CONFIG = "CROMWELL_CONFIG"
ENV_CROMWELL_ARGS = "CROMWELL_ARGS"
ENV_JAVA_HOME = "JAVA_HOME"
UNSAFE_RE = re.compile(r"[^\w.-]")


class Failures:
    def __init__(
        self,
        num_failed: int,
        failed_task: str,
        failed_task_exit_status: Optional[str] = None,
        failed_task_stdout: Optional[Union[Path, str]] = None,
        failed_task_stderr: Optional[Union[Path, str]] = None,
    ):
        self.num_failed = num_failed
        self.failed_task = failed_task
        self.failed_task_exit_status = failed_task_exit_status
        self._failed_task_stdout_path = None
        self._failed_task_stdout = None
        self._failed_task_stderr_path = None
        self._failed_task_stderr = None
        if isinstance(failed_task_stdout, Path):
            self._failed_task_stdout_path = cast(Path, failed_task_stdout)
        else:
            self._failed_task_stdout = cast(str, failed_task_stdout)
        if isinstance(failed_task_stderr, Path):
            self._failed_task_stderr_path = cast(Path, failed_task_stderr)
        else:
            self._failed_task_stderr = cast(str, failed_task_stderr)

    @property
    def failed_task_stdout(self):
        if self._failed_task_stdout is None and self._failed_task_stderr_path:
            self._failed_task_stdout = Failures._read_task_std(
                self._failed_task_stdout_path
            )
        return self._failed_task_stdout

    @property
    def failed_task_stderr(self):
        if self._failed_task_stderr is None and self._failed_task_stderr_path:
            self._failed_task_stderr = Failures._read_task_std(
                self._failed_task_stderr_path
            )
        return self._failed_task_stderr

    @staticmethod
    def _read_task_std(path: Path) -> Optional[str]:
        if path:
            if not path.exists():
                path = path.with_suffix(".background")
            if path.exists():
                with open(path, "rt") as inp:
                    return inp.read()


class CromwellHelperMixin:
    """
    Mixin for performing common tasks for cromwell executors
    """

    @classmethod
    def _get_cromwell_outputs(cls, output) -> dict:
        lines = output.splitlines(keepends=False)

        if len(lines) < 2:
            raise Exception(f"Invalid Cromwell output: {output}")

        if lines[1].startswith("Usage"):
            # If the cromwell command is not valid, usage is printed and the
            # return code is 0 so it does not cause an exception above - we
            # have to catch it here.
            raise Exception("Invalid Cromwell command")

        start = None

        for i, line in enumerate(lines):
            if line == "{" and lines[i + 1].lstrip().startswith('"outputs":'):
                start = i
            elif line == "}" and start is not None:
                end = i
                break
        else:
            raise AssertionError("No outputs JSON found in Cromwell stdout")

        return json.loads("\n".join(lines[start:(end + 1)]))["outputs"]

    @classmethod
    def _parse_metadata_errors(cls, metadata, target=None, error_kwargs=None):
        failures = cls._get_failures(metadata)

        if failures:
            error_kwargs.update(
                {
                    "failed_task": failures.failed_task,
                    "failed_task_exit_status": failures.failed_task_exit_status,
                    "failed_task_stdout": failures.failed_task_stdout,
                    "failed_task_stderr": failures.failed_task_stderr,
                }
            )
            if failures.num_failed > 1:
                error_kwargs["msg"] = (
                    f"cromwell failed on {failures.num_failed} instances of "
                    f"{failures.failed_task}; only "
                    f"showing output from the first failed task"
                )
        else:
            error_kwargs["msg"] = f"cromwell failed on workflow {target}"

    @classmethod
    def _get_failures(cls, metadata: dict) -> Optional[Failures]:
        for call_name, call_metadatas in metadata["calls"].items():
            failed = list(
                filter(lambda md: md["executionStatus"] == "Failed", call_metadatas)
            )

            if failed:
                failed_call = failed[0]

                if "subWorkflowMetadata" in failed_call:
                    return cls._get_failures(failed_call["subWorkflowMetadata"])
                else:
                    if "returnCode" in failed_call:
                        rc = failed_call["returnCode"]
                    else:
                        rc = "Unknown"

                    stdout = stderr = None

                    if "stdout" in failed_call:
                        stdout = Path(failed_call["stdout"])
                        stderr = Path(failed_call["stderr"])
                    elif "failures" in failed_call:
                        failure = failed_call["failures"][0]
                        stderr = failure["message"]

                        if "causedBy" in failure:
                            stderr = "\n  ".join(
                                [stderr] + [cb["message"] for cb in failure["causedBy"]]
                            )

                    return Failures(len(failed), call_name, rc, stdout, stderr)

    @classmethod
    def _get_workflow_imports(
        cls,
        import_dirs: Optional[Sequence[Path]] = None,
        imports_file: Optional[Path] = None,
    ) -> Path:
        """
        Creates a ZIP file with all WDL files to be imported.

        Args:
            imports_file: Text file naming import directories/files - one per line.

        Returns:
            Path to the ZIP file.
        """
        write_imports = bool(import_dirs)
        imports_path = None

        if imports_file:
            imports_path = ensure_path(imports_file)

            if imports_path.exists():
                write_imports = False

        if write_imports and import_dirs:
            imports = [
                wdl for path in import_dirs for wdl in glob.glob(str(path / "*.wdl"))
            ]

            if imports:
                if imports_path:
                    ensure_path(imports_path, is_file=True, create=True)
                else:
                    imports_path = Path(tempfile.mkstemp(suffix=".zip")[1])

                imports_str = " ".join(imports)

                LOG.info(f"Writing imports {imports_str} to zip file {imports_path}")

                exe = subby.run(
                    f"zip -j - {imports_str}",
                    mode=bytes,
                    stdout=imports_path,
                    raise_on_error=False,
                )

                if not exe.ok:
                    raise Exception(
                        f"Error creating imports zip file; stdout={exe.output}; "
                        f"stderr={exe.error}"
                    )

        return imports_path



from abc import ABCMeta, abstractmethod
import json
from typing import Any, IO, Optional, Sequence, cast

from pytest_wdl.core import DataDirs, DataManager, DataResolver
from pytest_wdl.utils import ensure_path

from py.path import local
import pytest
from _pytest.fixtures import FixtureRequest

try:
    import yaml
except ImportError:
    yaml = None


def pytest_collection(session: pytest.Session):
    """
    Prints an empty line to make the report look slightly better.
    """
    print()


def pytest_collect_file(path: local, parent) -> Optional[pytest.File]:
    if path.basename.startswith("test") and not path.basename.startswith("test_data."):
        if path.ext == ".json":
            return JsonWdlTestsModule(path, parent)
        elif yaml and path.ext == ".yaml":
            return YamlWdlTestsModule(path, parent)


# TODO: the Node API will be changing at some point
# https://docs.pytest.org/en/latest/example/nonpython.html#a-basic-example-for-specifying-tests-in-yaml-files


class WdlTestsModule(pytest.Module, metaclass=ABCMeta):
    @abstractmethod
    def _load(self, fp: IO) -> dict:
        pass

    def collect(self):
        with self.fspath.open() as inp:
            d = self._load(inp)

        if "tests" not in d:
            raise ValueError(f"Tests file {self.fspath} must contain a 'tests' key")

        data = d.get("data")

        for spec in d["tests"]:
            if "name" not in spec:
                raise ValueError("Test case missing 'name' key")

            yield TestItem(self, data=data, **spec)


class YamlWdlTestsModule(WdlTestsModule):
    def _load(self, fp: IO) -> dict:
        return yaml.safe_load(fp)


class JsonWdlTestsModule(WdlTestsModule):
    def _load(self, fp: IO) -> dict:
        return json.load(fp)


class TestItem(pytest.Item):
    def __init__(
        self,
        parent,
        data: Optional[dict] = None,
        name: Optional[str] = None,
        wdl: Optional[str] = None,
        inputs: Optional[dict] = None,
        expected: Optional[dict] = None,
        tags: Optional[Sequence] = None,
        **kwargs
    ):
        if not all((name, wdl)):
            raise ValueError("Every test must have 'name' and 'wdl' keys")

        super().__init__(name, parent)
        self._wdl = wdl
        self._inputs = inputs
        self._expected = expected
        self._tags = tags  # TODO: add tags as marks
        self._workflow_runner_kwargs = kwargs
        self._data = data
        self._fixture_request = None

    def setup(self):
        """
        This method is black magic - uses internal pytest APIs to create a
        FixtureRequest that can be used to access fixtures in `runtest()`.
        Copied from
        https://github.com/pytest-dev/pytest/blob/master/src/_pytest/doctest.py.
        """
        def func():
            pass

        self.funcargs = {}
        fm = self.session._fixturemanager
        self._fixtureinfo = fm.getfixtureinfo(
            node=self, func=func, cls=None, funcargs=False
        )
        self._fixture_request = FixtureRequest(self)
        self._fixture_request._fillfixtures()

    def runtest(self):
        # Get/create DataManager
        if self._data:
            config = self._fixture_request.getfixturevalue("user_config")
            data_resolver = DataResolver(self._data, config)
            data_dirs = DataDirs(
                ensure_path(self._fixture_request.fspath.dirpath(), canonicalize=True),
                function=self.name,
                module=None,  # TODO: support a top-level key for module name
                cls=None,  # TODO: support test groupings
            )
            workflow_data = DataManager(data_resolver, data_dirs)
        else:
            workflow_data = self._fixture_request.getfixturevalue("workflow_data")

        # Build the arguments to workflow_runner
        workflow_runner_kwargs = self._workflow_runner_kwargs

        # Resolve test data requests in the inputs and outputs

        if self._inputs:
            workflow_runner_kwargs["inputs"] = _resolve_test_data(
                self._inputs, workflow_data
            )

        if self._expected:
            workflow_runner_kwargs["expected"] = _resolve_test_data(
                self._expected, workflow_data
            )

        # Run the test
        workflow_runner = self._fixture_request.getfixturevalue("workflow_runner")

        return workflow_runner(self._wdl, **workflow_runner_kwargs)


def _resolve_test_data(d: dict, workflow_data: DataManager) -> dict:
    def _resolve(val: Any):
        if isinstance(val, str):
            try:
                # See if it's a test data entry
                return workflow_data[cast(str, val)]
            except FileNotFoundError:
                # It's a string literal
                return val
        elif isinstance(val, dict):
            return dict((key, _resolve(value)) for key, value in cast(dict, d).items())
        elif isinstance(val, Sequence):
            return [_resolve(value) for value in cast(Sequence, val)]
        else:
            return val

    return _resolve(d)



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from abc import ABCMeta, abstractmethod
import json
import os
from pathlib import Path
import tempfile
import textwrap
from typing import Any, Optional, Sequence, Tuple, Union, cast

from pytest_wdl.data_types import DataFile
from pytest_wdl.utils import (
    ensure_path, safe_string, find_executable_path, find_in_classpath
)

from WDL import Document, Error, Tree


ENV_JAVA_HOME = "JAVA_HOME"
ENV_JAVA_ARGS = "JAVA_ARGS"
INDENT = " " * 16


class ExecutorError(Exception):
    def __init__(self, executor: str, msg: Optional[str] = None):
        super().__init__(msg)
        self.executor = executor


class ExecutionFailedError(ExecutorError):
    def __init__(
        self,
        executor: str,
        target: str,
        status: str,
        inputs: Optional[dict] = None,
        executor_stdout: Optional[str] = None,
        executor_stderr: Optional[str] = None,
        failed_task: Optional[str] = None,
        failed_task_exit_status: Optional[int] = None,
        failed_task_stdout: Optional[str] = None,
        failed_task_stderr: Optional[str] = None,
        msg: Optional[str] = None
    ):
        if msg is None:
            if failed_task and failed_task != target:
                msg = f"{executor} failed with status {status} while running task " \
                      f"{failed_task} of {target}"
            else:
                msg = f"{executor} failed with status {status} while running {target}"
        super().__init__(executor, msg)
        self.target = target
        self.status = status
        self.inputs = inputs
        self.executor_stdout = executor_stdout
        self.executor_stderr = executor_stderr
        self.failed_task = failed_task
        self.failed_task_exit_status = failed_task_exit_status
        self.failed_task_stdout = failed_task_stdout
        self.failed_task_stderr = failed_task_stderr

    @property
    def exit_status_str(self) -> str:
        if self.failed_task_exit_status:
            return str(self.failed_task_exit_status)
        else:
            return "Unknown"

    def __str__(self):
        def wrap_std(std: str):
            if std:
                return "\n" + textwrap.indent(std, INDENT)
            else:
                return " None"

        return textwrap.dedent(f"""
        {self.args[0]}:
            inputs:
                {self.inputs}
            executor_stdout:{wrap_std(self.executor_stdout)}
            executor_stderr:{wrap_std(self.executor_stderr)}
            failed_task_exit_status: {self.exit_status_str}
            failed_task_stdout:{wrap_std(self.failed_task_stdout)}
            failed_task_stderr:{wrap_std(self.failed_task_stderr)}
        """)


class Executor(metaclass=ABCMeta):
    """
    Base class for WDL workflow executors.
    """

    @abstractmethod
    def run_workflow(
        self,
        wdl_path: Path,
        inputs: Optional[dict] = None,
        expected: Optional[dict] = None,
        **kwargs
    ) -> dict:
        """
        Run a WDL workflow on given inputs, and check that the output matches
        given expected values.

        Args:
            wdl_path: The WDL script to execute.
            inputs: Object that will be serialized to JSON and provided to Cromwell
                as the workflow inputs.
            expected: Dict mapping output parameter names to expected values.
            kwargs: Additional executor-specific keyword arguments (mostly for
                debugging)

        Returns:
            Dict of outputs.

        Raises:
            ExecutionFailedError: if there was an error executing the workflow
            AssertionError: if the actual outputs don't match the expected outputs
        """

    @classmethod
    def _validate_outputs(
        cls: "Executor", outputs: dict, expected: dict, target: str
    ) -> None:
        """
        Validate expected and actual outputs are equal.

        Args:
            outputs: Actual outputs
            expected: Expected outputs
            target: Execution target (i.e. workflow name)

        Raises:
            AssertionError
        """
        for name, expected_value in expected.items():
            key = f"{target}.{name}"
            if key not in outputs:
                raise AssertionError(f"Workflow did not generate output {key}")
            cls._compare_output_values(expected_value, outputs[key], key)

    @classmethod
    def _compare_output_values(
        cls: "Executor", expected_value, actual_value, name: str
    ) -> None:
        """
        Compare two values and raise an error if they are not equal.

        Args:
            expected_value:
            actual_value:
            name: Name of the output being compared

        Raises:
            AssertionError
        """
        if actual_value is None:
            if expected_value is None:
                return
            else:
                raise AssertionError(
                    f"Expected and actual values differ for {name}: "
                    f"{expected_value} != {actual_value}"
                )
        elif isinstance(expected_value, list):
            if len(expected_value) != len(actual_value):
                raise AssertionError(
                    f"Expected and actual values differ in length for {name}: "
                    f"{len(expected_value)} != {len(actual_value)}"
                )

            for i, (exp, act) in enumerate(zip(expected_value, actual_value)):
                cls._compare_output_values(exp, act, f"{name}[{i}]")
        elif isinstance(expected_value, dict):
            if len(expected_value) != len(actual_value):
                raise AssertionError(
                    f"Expected and actual values differ in length for {name}: "
                    f"{len(expected_value)} != {len(actual_value)}"
                )

            for key, exp in expected_value.items():
                if key not in actual_value:
                    raise AssertionError(
                        f"Key '{key}' is in the expected value but not in "
                        f"the actual value: {expected_value} != {actual_value}"
                    )

                cls._compare_output_values(exp, actual_value[key], f"{name}.{key}")
        elif isinstance(expected_value, DataFile):
            # TODO: pass name
            expected_value.assert_contents_equal(actual_value)
        elif expected_value != actual_value:
            raise AssertionError(
                f"Expected and actual values differ for {name}: "
                f"{expected_value} != {actual_value}"
            )


class JavaExecutor(Executor, metaclass=ABCMeta):
    """
    Manages the running of WDL workflows using a Java-based executor.

    Args:
        java_bin: Path to the java executable.
        java_args: Default Java arguments to use; can be overidden by passing
            `java_args=...` to `run_workflow`.
    """

    def __init__(
        self,
        java_bin: Optional[Union[str, Path]] = None,
        java_args: Optional[str] = None
    ):
        if not java_bin:
            java_home = os.environ.get(ENV_JAVA_HOME)
            if java_home:
                java_bin = Path(java_home) / "bin" / "java"
            else:
                java_bin = find_executable_path("java")

        if not java_bin:
            raise FileNotFoundError("Could not find java executable")

        self.java_bin = ensure_path(
            java_bin, exists=True, is_file=True, executable=True
        )

        self.java_args = java_args or os.environ.get(ENV_JAVA_ARGS)

    @staticmethod
    def resolve_jar_file(
        file_name_pattern: str, jar_path: Optional[Path] = None,
        env_var: Optional[str] = None,
    ):
        if not jar_path:
            path_str = None
            if env_var:
                path_str = os.environ.get(env_var)
            if path_str:
                jar_path = ensure_path(path_str)
            else:
                jar_path = find_in_classpath(file_name_pattern)

        if not jar_path:
            raise FileNotFoundError(f"Could not find JAR file {file_name_pattern}")

        return ensure_path(jar_path, is_file=True, exists=True)


class InputsFormatter:
    @classmethod
    def get_instance(cls) -> "InputsFormatter":
        if not hasattr(cls, "__instance"):
            setattr(cls, "__instance", object.__new__(cls))

        return getattr(cls, "__instance")

    def format_inputs(
        self, inputs_dict: dict, namespace: Optional[str] = None
    ) -> dict:
        prefix = f"{namespace}." if namespace else ""
        return dict(
            (f"{prefix}{key}", self.format_value(value))
            for key, value in inputs_dict.items()
        )

    def format_value(self, value: Any) -> Any:
        """
        Convert a primitive, DataFile, Sequence, or Dict to a JSON-serializable object.
        Currently, arbitrary objects can be serialized by implementing an `as_dict()`
        method, otherwise they are converted to strings.

        Args:
            value: The value to format.

        Returns:
            The serializable value.
        """
        if hasattr(value, "as_dict"):
            return value.as_dict()

        if isinstance(value, DataFile):
            return self._format_data_file(cast(DataFile, value))

        if isinstance(value, dict):
            return self._format_dict(cast(dict, value))

        if isinstance(value, Sequence) and not isinstance(value, str):
            return self._format_sequence(cast(Sequence, value))

        return value

    def _format_sequence(self, s: Sequence) -> list:
        return [self.format_value(val) for val in s]

    def _format_dict(self, d: dict) -> dict:
        return dict((key, self.format_value(val)) for key, val in d.items())

    def _format_data_file(self, df: DataFile) -> Union[str, dict]:
        return df.path


def parse_wdl(
    wdl_path: Path,
    import_dirs: Optional[Sequence[Path]] = (),
    check_quant: bool = False,
    **_
) -> Document:
    return Tree.load(
        str(wdl_path),
        path=[str(path) for path in import_dirs],
        check_quant=check_quant
    )


def get_target_name(
    wdl_path: Optional[Path] = None,
    wdl_doc: Optional[Document] = None,
    task_name: Optional[str] = None,
    workflow_name: Optional[str] = None,
    **kwargs
) -> Tuple[str, bool]:
    """
    Get the execution target. The order of priority is:

    - task_name
    - workflow_name
    - wdl_doc.workflow.name
    - wdl_doc.task[0].name
    - wdl_file.stem

    Args:
        wdl_path: Path to a WDL file
        wdl_doc: A miniwdl-parsed WDL document
        task_name: The task name
        workflow_name: The workflow name
        **kwargs: Additional keyword arguments to pass to `parse_wdl`

    Returns:
        A tuple (target, is_task), where `is_task` is a boolean indicating whether
        the target is a task (True) or a workflow (False).

    Raises:
        ValueError if 1) neither `task_name` nor `workflow_name` is specified and the
        WDL document contains no workflow and multiple tasks; or 2) all of the
        parameters are None.
    """
    if task_name:
        return task_name, True

    if workflow_name:
        return workflow_name, False

    if not wdl_doc and Tree:
        try:
            wdl_doc = parse_wdl(wdl_path, **kwargs)
        except Error.SyntaxError as err:
            raise RuntimeError(
                "There was an error parsing the WDL document to extract the target "
                "workflow/task name. Please specify the 'workflow_name' or 'task_name' "
                "parameter to workflow_runner()."
            ) from err

    if wdl_doc:
        if wdl_doc.workflow:
            return wdl_doc.workflow.name, False
        elif wdl_doc.tasks and len(wdl_doc.tasks) == 1:
            return wdl_doc.tasks[0].name, True
        else:
            raise ValueError(
                "WDL document has no workflow and multiple tasks, and 'task_name' "
                "is not specified"
            )

    if wdl_path:
        return safe_string(wdl_path.stem), False

    raise ValueError("At least one parameter must not be None")


def read_write_inputs(
    inputs_file: Optional[Union[str, Path]] = None,
    inputs_dict: Optional[dict] = None,
    inputs_formatter: Optional[InputsFormatter] = InputsFormatter.get_instance(),
    write_formatted_inputs: bool = True,
    **kwargs
) -> Tuple[dict, Optional[Path]]:
    """
    If `inputs_file` is specified and it exists, read its contents. Otherwise, if
    `inputs_dict` is specified, format it using `inputs_formatter` (if specified) and
    write it to `inputs_file` or a temporary file.

    Args:
        inputs_file:
        inputs_dict:
        inputs_formatter:
        write_formatted_inputs:
        kwargs:

    Returns:
        The (formatted) inputs dict and the resolved inputs file. If both `inputs_dict`
        and `inputs_file` are None, returns `({}, None)`.
    """
    if inputs_file:
        inputs_file = ensure_path(inputs_file, is_file=True, create=True)

        if inputs_file.exists():
            with open(inputs_file, "rt") as inp:
                inputs_dict_from_file = json.load(inp)
                return inputs_dict_from_file, inputs_file

    if inputs_dict:
        inputs_dict = inputs_formatter.format_inputs(inputs_dict, **kwargs)

        if write_formatted_inputs:
            if not inputs_file:
                inputs_file = Path(tempfile.mkstemp(suffix=".json")[1])

            with open(inputs_file, "wt") as out:
                json.dump(inputs_dict, out, default=str)

        return inputs_dict, inputs_file

    return {}, None



#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import logging
from pathlib import Path
from typing import Optional, Sequence, cast

from pytest_wdl.executors import (
    Executor, ExecutionFailedError, get_target_name, read_write_inputs
)

from WDL import CLI, Error, Tree, runtime


class MiniwdlExecutor(Executor):
    """
    Manages the running of WDL workflows using Cromwell.
    """

    def __init__(self, import_dirs: Optional[Sequence[Path]] = None):
        self._import_dirs = import_dirs

    def run_workflow(
        self,
        wdl_path: Path,
        inputs: Optional[dict] = None,
        expected: Optional[dict] = None,
        **kwargs
    ) -> dict:
        """
        Run a WDL workflow on given inputs, and check that the output matches
        given expected values.

        Args:
            wdl_path: The WDL script to execute.
            inputs: Object that will be serialized to JSON and provided to Cromwell
                as the workflow inputs.
            expected: Dict mapping output parameter names to expected values.
            kwargs: Additional keyword arguments, mostly for debugging:
                * workflow_name: Name of the workflow to run.
                * task_name: Name of the task to run if a workflow isn't defined.
                * inputs_file: Path to the Cromwell inputs file to use. Inputs are
                    written to this file only if it doesn't exist.

        Returns:
            Dict of outputs.

        Raises:
            Exception: if there was an error executing Cromwell
            AssertionError: if the actual outputs don't match the expected outputs
        """

        logger = logging.getLogger("miniwdl-run")
        logger.setLevel(CLI.NOTICE_LEVEL)
        CLI.install_coloredlogs(logger)

        wdl_doc = CLI.load(
            str(wdl_path),
            path=[str(path) for path in self._import_dirs],
            check_quant=kwargs.get("check_quant", True),
            read_source=CLI.read_source
        )

        namespace, is_task = get_target_name(wdl_doc=wdl_doc, **kwargs)

        inputs_dict, inputs_file = read_write_inputs(
            inputs_dict=inputs, namespace=namespace if not is_task else None,
        )

        target, input_env, input_json = CLI.runner_input(
            doc=wdl_doc,
            inputs=[],
            input_file=str(inputs_file) if inputs_file else None,
            empty=[],
            task=namespace if is_task else None
        )

        # Create config
        cfg = runtime.config.Loader(logger)
        cfg.override({
            "copy_input_files": kwargs.get("copy_input_files", False)
        })
        cfg.log_all()

        # initialize Docker
        runtime.task.LocalSwarmContainer.global_init(cfg, logger)

        try:
            rundir, output_env = runtime.run(cfg, target, input_env, run_dir=None)
        except Error.EvalError as err:  # TODO: test errors
            MiniwdlExecutor.log_source(logger, err)
            raise
        except Error.RuntimeError as err:
            MiniwdlExecutor.log_source(logger, err)

            if isinstance(err, runtime.error.RunFailed):
                # This will be a workflow- or a task-level failure, depending on
                # whether a workflow or task was executed. If it is workflow-level,
                # we need to get the task-level error that caused the workflow to fail.
                if isinstance(err.exe, Tree.Workflow):
                    err = err.__cause__

                task_err = cast(runtime.error.RunFailed, err)
                cause = task_err.__cause__
                failed_task_exit_status = None
                failed_task_stderr = None
                if isinstance(cause, runtime.error.CommandFailed):
                    # If the task failed due to an error in the command, populate the
                    # command exit status and stderr.
                    cmd_err = cast(runtime.error.CommandFailed, cause)
                    failed_task_exit_status = cmd_err.exit_status
                    failed_task_stderr = MiniwdlExecutor.read_miniwdl_command_std(
                        cmd_err.stderr_file
                    )

                raise ExecutionFailedError(
                    "miniwdl",
                    namespace,
                    status="Failed",
                    inputs=task_err.exe.inputs,
                    failed_task=task_err.exe.name,
                    failed_task_exit_status=failed_task_exit_status,
                    failed_task_stderr=failed_task_stderr
                ) from err
            else:
                raise

        outputs = CLI.values_to_json(output_env, namespace=target.name)

        if expected:
            self._validate_outputs(outputs, expected, target.name)

        return outputs

    @staticmethod
    def read_miniwdl_command_std(path: Optional[str] = None) -> Optional[str]:
        if path:
            p = Path(path)

            if p.exists():
                with open(path, "rt") as inp:
                    return inp.read()

    @staticmethod
    def log_source(logger: logging.Logger, exn: Exception):
        if isinstance(exn, runtime.error.RunFailed):
            pos = cast(runtime.error.RunFailed, exn).exe.pos
        elif hasattr(exn, "pos"):
            pos = cast(Error.SourcePosition, getattr(exn, "pos"))
        else:
            return
        if pos:
            logger.error(
                "({} Ln {} Col {}) {}{}".format(
                    pos.uri,
                    pos.line,
                    pos.column,
                    exn.__class__.__name__,
                    (", " + str(exn) if str(exn) else ""),
                )
            )



#! /usr/bin/env python
#
#    Copyright 2019 Eli Lilly and Company
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# TODO: some of the code here can be replaced by functions in xphyle.{paths,utils}
import contextlib
import fnmatch
import hashlib
import logging
import os
from pathlib import Path
import re
import shutil
import stat
import tempfile
import time
from typing import Callable, Optional, Sequence, Union, cast

from py._path.local import LocalPath


LOG = logging.getLogger("pytest-wdl")
LOG.setLevel(os.environ.get("LOGLEVEL", "WARNING").upper())

ENV_PATH = "PATH"
ENV_CLASSPATH = "CLASSPATH"
DEFAULT_CLASSPATH = "."

UNSAFE_RE = re.compile(r"[^\w.-]")


def safe_string(s: str, replacement: str = "_") -> str:
    """
    Makes a string safe by replacing non-word characters.

    Args:
        s: The string to make safe
        replacement: The replacement stringj

    Returns:
        The safe string
    """
    return UNSAFE_RE.sub(replacement, s)


# def deprecated(f: Callable):
#     """
#     Decorator for deprecated functions/methods. Deprecated functionality will be
#     removed before each major release.
#     """
#     def decorator(*args, **kwargs):
#         LOG.warning(f"Function/method {f.__name__} is deprecated and will be removed")
#         f(*args, **kwargs)
#     return decorator


@contextlib.contextmanager
def chdir(todir: Path):
    """
    Context manager that temporarily changes directories.

    Args:
        todir: The directory to change to.
    """
    curdir = Path.cwd()
    try:
        os.chdir(todir)
        yield todir
    finally:
        os.chdir(curdir)


@contextlib.contextmanager
def tempdir(
    change_dir: bool = False,
    tmproot: Optional[Path] = None,
    cleanup: Optional[bool] = True,
) -> Path:
    """
    Context manager that creates a temporary directory, yields it, and then
    deletes it after return from the yield.

    Args:
        change_dir: Whether to temporarily change to the temp dir.
        tmproot: Root directory in which to create temporary directories.
        cleanup: Whether to delete the temporary directory before exiting the context.
    """
    temp = ensure_path(tempfile.mkdtemp(dir=tmproot))
    try:
        if change_dir:
            with chdir(temp):
                yield temp
        else:
            yield temp
    finally:
        if cleanup:
            shutil.rmtree(temp)


@contextlib.contextmanager
def context_dir(
    path: Optional[Path] = None,
    change_dir: bool = False,
    cleanup: Optional[bool] = None,
) -> Path:
    """
    Context manager that looks for a specific environment variable to specify a
    directory. If the environment variable is not set, a temporary directory is
    created and cleaned up upon return from the yield.

    Args:
        path: The environment variable to look for.
        change_dir: Whether to change to the directory.
        cleanup: Whether to delete the directory when exiting the context. If None,
            the directory is only deleted if a temporary directory is created.

    Yields:
        A directory path.
    """
    if cleanup is None:
        cleanup = path is None

    if not path:
        path = Path(tempfile.mkdtemp())
    elif not path.exists():
        path.mkdir(parents=True)

    try:
        if change_dir:
            with chdir(path):
                yield path
        else:
            yield path
    finally:
        if cleanup and path.exists():
            shutil.rmtree(path, ignore_errors=True)


def ensure_path(
    path: Union[str, LocalPath, Path],
    search_paths: Optional[Sequence[Path]] = None,
    canonicalize: bool = True,
    exists: Optional[bool] = None,
    is_file: Optional[bool] = None,
    executable: Optional[bool] = None,
    create: bool = False,
) -> Path:
    """
    Converts a string path or :class:`py.path.local.LocalPath` to a
    :class:`pathlib.Path`.

    Args:
        path: The path to convert.
        search_paths: Directories to search for `path` if it is not already absolute.
            If `exists` is True, looks for the first search path that contains the file,
            otherwise just uses the first search path.
        canonicalize: Whether to return the canonicalized version of the path -
            expand home directory shortcut (~), make absolute, and resolve symlinks.
        exists: If True, raise an exception if the path does not exist; if False,
            raise an exception if the path does exist.
        is_file: If True, raise an exception if the path is not a file; if False,
            raise an exception if the path is not a directory.
        executable: If True and `is_file` is True and the file exists, raise an
            exception if it is not executable.
        create: Create the directory (or parent, if `is_file` = True) if
            it does not exist. Ignored if `exists` is True.

    Returns:
        A `pathlib.Path` object.
    """
    if isinstance(path, Path):
        p = cast(Path, path)
    else:
        p = Path(str(path))

    p = Path(os.path.expandvars(p))

    if canonicalize:
        p = p.expanduser()

        if search_paths and not p.is_absolute():
            if exists:
                for search_path in search_paths:
                    p_tmp = search_path / p
                    if p_tmp.exists():
                        p = p_tmp.absolute()
                        break
            else:
                p = (search_paths[0] / p).absolute()

        p = p.resolve()

    if p.exists():
        if exists is False:
            raise FileExistsError(f"Path {p} already exists")
        if is_file is True:
            if p.is_dir():
                raise IsADirectoryError(f"Path {p} is not a file")
            elif executable and not is_executable(p):
                raise OSError(f"File {p} is not executable")
        elif is_file is False and not p.is_dir():
            raise NotADirectoryError(f"Path {p} is not a directory")
    elif exists is True:
        raise FileNotFoundError(f"Path {p} does not exist")
    elif create:
        if is_file:
            p.parent.mkdir(parents=True, exist_ok=True)
        else:
            p.mkdir(parents=True, exist_ok=True)

    return p


def resolve_file(
    filename: Union[str, Path], project_root: Path, assert_exists: bool = True
) -> Optional[Path]:
    """
    Finds `filename` under `project_root` or in the project path.

    Args:
        filename: The filename, relative path, or absolute path to resolve.
        project_root: The project root dir.
        assert_exists: Whether to raise an error if the file cannot be found.

    Returns:
        A `pathlib.Path` object, or None if the file cannot be found and
        `assert_exists` is False.

    Raises:
        FileNotFoundError if the file cannot be found and `assert_exists` is True.
    """
    path = ensure_path(filename, canonicalize=False)
    is_abs = path.is_absolute()

    if is_abs and path.exists():
        return path

    if not is_abs:
        check_path = ensure_path(project_root / path)
        if check_path.exists():
            return check_path
        # Search in cwd
        check_path = find_project_path(path)
        if check_path and check_path.exists():
            return check_path
        # Search upward from project root
        check_path = find_project_path(path, start=project_root)
        if check_path and check_path.exists():
            return check_path

    if assert_exists:
        raise FileNotFoundError(f"Could not resolve file: {filename}")
    else:
        return None


def find_project_path(
    *filenames: Union[str, Path],
    start: Optional[Path] = None,
    return_parent: bool = False,
    assert_exists: bool = False,
) -> Optional[Path]:
    """
    Starting from `path` folder and moving upwards, search for any of `filenames` and
    return the first path containing any one of them.

    Args:
        *filenames: Filenames to search. Either a string filename, or a sequence of
            string path elements.
        start: Starting folder
        return_parent: Whether to return the containing folder or the discovered file.
        assert_exists: Whether to raise an exception if a file cannot be found.

    Returns:
        A `Path`, or `None` if no folder is found that contains any of `filenames`.
        If `return_parent` is `False` and more than one of the files is found one
        of the files is randomly selected for return.

    Raises:
        FileNotFoundError if the file cannot be found and `assert_exists` is True.
    """
    path = start or Path.cwd()
    while path != path.parent:
        for filename in filenames:
            if isinstance(filename, str):
                found = list(path.glob(filename))
                found = found[0] if found else None
            else:
                found = path / filename
                if not found.exists():
                    found = None
            if found:
                LOG.debug("Found %s in %s", filename, path)
                if return_parent:
                    return path
                else:
                    return found
        else:
            path = path.parent

    if assert_exists:
        raise FileNotFoundError(
            f"Could not find any of {','.join(str(f) for f in filenames)} "
            f"starting from {start}"
        )

    return None


def find_executable_path(
    executable: str, search_path: Optional[Sequence[Path]] = None
) -> Optional[Path]:
    """Finds 'executable' in `search_path`.

    Args:
        executable: The name of the executable to find.
        search_path: The list of directories to search. If None, the system search
            path (defined by the $PATH environment variable) is used.

    Returns:
        Absolute path of the executable, or None if no matching executable was found.
    """
    if search_path is None:
        if ENV_PATH in os.environ:
            search_path = [Path(p) for p in os.environ[ENV_PATH].split(os.pathsep)]
        else:
            return None
    for path in search_path:
        exe_path = path / executable
        if exe_path.exists() and is_executable(exe_path):
            return exe_path
    else:
        return None


def is_executable(path: Path) -> bool:
    """
    Checks if a path is executable.

    Args:
        path: The path to check

    Returns:
        True if `path` exists and is executable by the user, otherwise False.
    """
    return path.exists() and os.stat(path).st_mode & stat.S_IXUSR


def find_in_classpath(glob: str) -> Optional[Path]:
    """
    Attempts to find a .jar file matching the specified glob pattern in the
    Java classpath.

    Args:
        glob: JAR filename pattern

    Returns:
        Path to the JAR file, or None if a matching file is not found.
    """
    classpath = os.environ.get(ENV_CLASSPATH, DEFAULT_CLASSPATH)

    for path_str in classpath.split(os.pathsep):
        path = ensure_path(path_str)
        if path.exists():
            if path.is_dir():
                matches = list(path.glob(glob))
                if matches:
                    if len(matches) > 1:
                        LOG.warning(
                            "Found multiple jar files matching pattern %s: %s;"
                            "returning the first one.",
                            glob,
                            matches,
                        )
                    return matches[0]
            elif path.exists() and fnmatch.fnmatch(path.name, glob):
                return path


def env_map(d: dict) -> dict:
    """
    Given a mapping of keys to value descriptors, creates a mapping of the keys to
    the described values.
    """
    envmap = {}
    for name, value_descriptor in d.items():
        value = resolve_value_descriptor(value_descriptor)
        if value:
            envmap[name] = value
    return envmap


def resolve_value_descriptor(value_descriptor: Union[str, dict]) -> Optional:
    """
    Resolves the value of a value descriptor, which may be an environment variable
    name, or a map with keys `env` (the environment variable name) and `value` (the
    value to use if `env` is not specified or if the environment variable is unset.

    Args:
        value_descriptor:

    Returns:

    """
    if isinstance(value_descriptor, str):
        return os.environ.get(value_descriptor)
    elif "env" in value_descriptor:
        return os.environ.get(value_descriptor["env"], value_descriptor.get("value"))
    else:
        return value_descriptor.get("value")


class DigestsNotEqualError(AssertionError):
    pass


def compare_files_with_hash(file1: Path, file2: Path, hash_name: str = "md5"):
    file1_digest = hash_file(file1, hash_name)
    file2_digest = hash_file(file2, hash_name)
    if file1_digest != file2_digest:
        raise DigestsNotEqualError(
            f"{hash_name} digests differ between expected identical files "
            f"{file1}, {file2}"
        )


def hash_file(path: Path, hash_name: str = "md5") -> str:
    assert hash_name in hashlib.algorithms_guaranteed
    with open(path, "rb") as inp:
        hashobj = hashlib.new(hash_name)
        hashobj.update(inp.read())
        return hashobj.hexdigest()


def verify_digests(path: Path, digests: dict):
    for hash_name, expected_digest in digests.items():
        try:
            actual_digest = hash_file(path, hash_name)
        except AssertionError:  # TODO: test this
            LOG.warning(
                "Hash algorithm %s is not supported; cannot verify file %s",
                hash_name,
                path,
            )
            continue
        if actual_digest != expected_digest:
            raise DigestsNotEqualError(
                f"{hash_name} digest {actual_digest} of file "
                f"{path} does match expected value {expected_digest}"
            )


class PollingException(Exception):
    """Base exception that stores the last result seen."""
    def __init__(self, last=None):
        self.last = last


class TimeoutException(PollingException):
    """Exception raised if polling function times out"""


class MaxCallException(PollingException):
    """Exception raised if maximum number of iterations is exceeded"""


def poll(
    target: Callable,
    step: int = 1,
    args: Optional[Sequence] = None,
    kwargs: Optional[dict] = None,
    timeout: Optional[int] = None,
    max_tries: Optional[int] = None,
    check_success: Callable = bool,
    step_function: Optional[Callable[[int, int], int]] = None,
    ignore_exceptions: Sequence = (),
):
    """
    Poll by calling a target function until a certain condition is met. You must specify
    at least a target function to be called and the step -- base wait time between
    each function call.

    Vendored from the [polling](https://github.com/justiniso/polling) package.

    Args:
        target: The target callable
        step: Step defines the amount of time to wait (in seconds)
        args: Arguments to be passed to the target function
        kwargs: Keyword arguments to be passed to the target function
        timeout: The target function will be called until the time elapsed is greater
            than the maximum timeout (in seconds). NOTE that the actual execution
            time of the function *can* exceed the time specified in the timeout. For
            instance, if the target function takes 10 seconds to execute and the timeout
            is 21 seconds, the polling function will take a total of 30 seconds (two
            iterations of the target --20s which is less than the timeout--21s,
            and a final iteration)
        max_tries: Maximum number of times the target function will be called before
            failing
        check_success: A callback function that accepts the return value of the target
            function. It must return true if you want the polling function to stop
            and return this value. It must return false if you want to continue
            polling. You may also use this function to collect non-success values. The
            default is a callback that tests for truthiness (anything not False, 0,
            or empty collection).
        step_function: A callback function that accepts two arguments: current_step,
            num_tries; and returns the next step value. By default, this is constant,
            but you can also pass a function that will increase or decrease the step.
            As an example, you can increase the wait time between calling the target
            function by 10 seconds every iteration until the step is 100 seconds--at
            which point it should remain constant at 100 seconds

            >>> def my_step_function(current_step: int, num_tries: int) -> int:
            >>>     return max(current_step + 10, 100)

        ignore_exceptions: You can specify a tuple of exceptions that should be caught
            and ignored on every iteration. If the target function raises one of
            these exceptions, it will be caught and the exception instance will be
            pushed to the queue of values collected during polling. Any other exceptions
            raised will be raised as normal.

    Returns:
        The first value from the target function that meets the condions of the
        check_success callback. By default, this will be the first value that is not
        None, 0, False, '', or an empty collection.
    """
    max_time = time.time() + timeout if timeout else None
    tries = 0
    last_item = None

    if args is None:
        args = ()

    if kwargs is None:
        kwargs = {}

    while True:
        if max_tries and tries >= max_tries:
            raise MaxCallException(last_item)

        try:
            val = target(*args, **kwargs)
            last_item = val
        except ignore_exceptions as e:
            last_item = e
        else:
            # Condition passes, this is the only "successful" exit from the
            # polling function
            if check_success(val):
                return val

        tries += 1

        # Check the time after to make sure the poll function is called at least once
        if max_time and time.time() >= max_time:
            raise TimeoutException(last_item)

        time.sleep(step)

        if step_function:
            step = step_function(step, tries)



#! /usr/bin/env python

"""Test hello_world parent workflow"""
import pytest


@pytest.mark.integration
def test_workflow(workflow_data, workflow_runner):
    """Test the hello_world parent workflow with fixed inputs and outputs."""
    inputs = {
        "input_files": [
            workflow_data["test_file"],
            workflow_data["test_file"],
        ]
    }
    expected = {
        "single_file": workflow_data["test_file"]
    }
    workflow_runner(
        "workflow.wdl",
        inputs,
        expected,
        executors=["cromwell"]
    )



# Copyright 2019 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from argparse import Namespace
import contextlib
import json
import os
from pathlib import Path
import shutil
import tempfile
from typing import Any, Dict, Iterable, Optional, Sequence, Tuple, Union, cast
from unittest.mock import patch

from pytest_wdl import config
from pytest_wdl.data_types import DataFile
from pytest_wdl.executors import (
    ExecutorError,
    ExecutionFailedError,
    JavaExecutor,
    parse_wdl,
)
from pytest_wdl.localizers import UrlLocalizer
from pytest_wdl.plugins import PluginError
from pytest_wdl.url_schemes import Method, Request, Response, UrlHandler
from pytest_wdl.utils import LOG, ensure_path, verify_digests

try:
    # test whether dxpy is installed and the user is logged in
    import dxpy
except ImportError as err:
    raise PluginError(
        "DNAnexus (dx) extensions require that you install 'dxpy' "
        "(try 'pip install dxpy')."
    ) from err

from dxpy.scripts import dx
from dxpy.utils.job_log_client import DXJobLogStreamClient
import subby
from WDL import Document, Type


ENV_JAVA_HOME = "JAVA_HOME"
ENV_DXWDL_JAR = "DXWDL_JAR"
ENV_DX_USERNAME = "DX_USERNAME"
OUTPUT_STAGE = "stage-outputs"
DX_LINK_KEY = "$dnanexus_link"
DX_DICT_KEY = "___"
DX_FILES_SUFFIX = "___dxfiles"
STDOUT_LOG = "STDOUT"
STDERR_LOG = "STDERR"


@contextlib.contextmanager
def login(logout: bool = False, interactive: bool = True):
    """
    Checks that the user is logged into DNAnexus, otherwise log them in.

    Args:
        logout: Whether to log out before exiting the context. Ignored if the user is
            already logged in.
        interactive: Whether to allow interactive login.

    Raises:
        PluginError: if `interactive` is `False` and the user cannot be logged in
            non-interactively
    """
    if dxpy.SECURITY_CONTEXT:
        try:
            dxpy.whoami()
        except dxpy.exceptions.InvalidAuthentication:
            dxpy.SECURITY_CONTEXT = None

    if dxpy.SECURITY_CONTEXT:
        yield
    else:
        if config.get_instance():
            conf = config.get_instance().get_provider_defaults("dxwdl")
        else:
            conf = {}

        username = conf.get("username")
        token = conf.get("token")

        if ENV_DX_USERNAME not in os.environ and username:
            os.environ[ENV_DX_USERNAME] = username

        args = Namespace(
            auth_token=None,
            token=token,
            host=None,
            port=None,
            protocol="https",
            timeout=conf.get("timeout", "1d"),
            save=True,
            staging=False,
            projects=False
        )

        try:
            if not token and "password" in conf:
                # Use configured credentials to log user in automatically
                with patch("builtins.input", return_value=username), \
                        patch("getpass.getpass", return_value=conf["password"]):
                    dx.login(args)
            elif not (token or interactive):
                raise ValueError(
                    "User is not logged in, credentials were not provided, and "
                    "interactive login is not allowed"
                )
            else:
                # If token is not specified, this will require interactive login
                dx.login(args)

            yield
        except Exception as lerr:
            raise PluginError(
                "DNAnexus (dx) extensions require you to be logged into your account"
            ) from lerr
        finally:
            if logout:
                dx.logout(args)


class DxResponse(Response):
    def __init__(self, file_id: str, project_id: Optional[str] = None):
        self.file_id = file_id
        self.project_id = project_id

    def download_file(
        self,
        destination: Path,
        show_progress: bool = False,
        digests: Optional[dict] = None
    ):
        destination.parent.mkdir(parents=True, exist_ok=True)

        with login():
            dxpy.download_dxfile(
                self.file_id,
                str(destination),
                show_progress=show_progress,
                project=self.project_id
            )

        if digests:
            verify_digests(destination, digests)


class DxUrlHandler(UrlHandler):
    @property
    def scheme(self) -> str:
        return "dx"

    @property
    def handles(self) -> Sequence[Method]:
        return [Method.OPEN]

    def urlopen(self, request: Request) -> Response:
        url = request.get_full_url()

        if not url.startswith("dx://"):  # TODO: test this
            raise ValueError(f"Expected URL to start with 'dx://'; got {url}")

        obj_id = url[5:]

        if ":" in obj_id:
            project_id, file_id = obj_id.split(":")
        else:
            project_id = None
            file_id = obj_id

        return DxResponse(file_id, project_id)


class DxWdlExecutor(JavaExecutor):
    # TODO: not thread safe - tests cannot be parallelized
    _workflow_cache: Dict[Path, dxpy.DXWorkflow] = {}
    _data_cache = {}

    def __init__(
        self,
        import_dirs: Optional[Sequence[Path]] = None,
        java_bin: Optional[Union[str, Path]] = None,
        java_args: Optional[str] = None,
        dxwdl_jar_file: Optional[Union[str, Path]] = None,
        dxwdl_cache_dir: Optional[Union[str, Path]] = None,
    ):
        super().__init__(java_bin, java_args)
        self._import_dirs = import_dirs
        self._dxwdl_jar_file = self.resolve_jar_file(
            "dxWDL*.jar", dxwdl_jar_file, ENV_DXWDL_JAR
        )
        if dxwdl_cache_dir:
            self._dxwdl_cache_dir = ensure_path(dxwdl_cache_dir)
            self._cleanup_cache = False
        else:
            self._dxwdl_cache_dir = ensure_path(tempfile.mkdtemp())
            self._cleanup_cache = True

    def run_workflow(
        self,
        wdl_path: Path,
        inputs: Optional[dict] = None,
        expected: Optional[dict] = None,
        **kwargs
    ) -> dict:
        # TODO: handle "task_name" kwarg - run app instead of workflow
        wdl_doc = parse_wdl(wdl_path, self._import_dirs, **kwargs)

        if not wdl_doc.workflow:
            raise ValueError(
                "Currently, the dxWDL executor only supports executing "
                "workflows, not individual tasks"
            )

        workflow_name = wdl_doc.workflow.name

        if (
            "workflow_name" in kwargs
            and workflow_name != kwargs["workflow-name"]
        ):
            raise ValueError(
                f"The workflow name '{workflow_name}' does not match the value "
                f"of the 'workflow_name' parameter '{kwargs['workflow-name']}'"
            )

        namespace = kwargs.get("stage_id", "stage-common")
        inputs_dict = None

        if "inputs_file" in kwargs:
            inputs_file = ensure_path(kwargs["inputs_file"])

            if inputs_file.exists():
                with open(inputs_file, "rt") as inp:
                    inputs_dict = json.load(inp)

        if not inputs_dict:
            workflow_inputs = wdl_doc.workflow.available_inputs

            if workflow_inputs:
                dx_inputs_formatter = DxInputsFormatter(wdl_doc, **kwargs)
                inputs_dict = dx_inputs_formatter.format_inputs(inputs, namespace)
            else:
                inputs_dict = {}

        try:
            with login():
                workflow = self._resolve_workflow(wdl_path, workflow_name, kwargs)
                analysis = workflow.run(inputs_dict)

                try:
                    analysis.wait_on_done()

                    outputs = self._get_analysis_outputs(analysis, expected.keys())

                    if expected:
                        self._validate_outputs(outputs, expected, OUTPUT_STAGE)

                    return outputs
                except dxpy.exceptions.DXJobFailureError:
                    raise ExecutionFailedError(
                        "dxWDL",
                        workflow_name,
                        analysis.describe()["state"],
                        inputs_dict,
                        **self._get_failed_task(analysis)
                    )
                finally:
                    if self._cleanup_cache:
                        shutil.rmtree(self._dxwdl_cache_dir)
        except dxpy.exceptions.InvalidAuthentication as ierr:
            raise ExecutorError("dxwdl", "Invalid DNAnexus credentials/token") from ierr
        except dxpy.exceptions.ResourceNotFound as rerr:
            raise ExecutorError("dxwdl", "Required resource was not found") from rerr
        except dxpy.exceptions.PermissionDenied as perr:
            raise ExecutorError(
                "dxwdl", f"You must have at least CONTRIBUTE permission"
            ) from perr

    def _resolve_workflow(
        self,
        wdl_path: Path,
        workflow_name: str,
        kwargs: dict,
    ) -> dxpy.DXWorkflow:
        if wdl_path in DxWdlExecutor._workflow_cache:
            return DxWdlExecutor._workflow_cache[wdl_path]

        project_id = (
            kwargs.get("workflow_project_id") or
            kwargs.get("project_id", dxpy.PROJECT_CONTEXT_ID)
        )
        folder = kwargs.get("workflow_folder") or kwargs.get("folder", "/")

        # # This probably isn't necessary, since (I think) dxWDL creates the folder
        # # if it doesn't exist
        # if not folder:
        #     folder = "/"
        # else:
        #     # Check that the project exists and create the folder (any any missing
        #     # parents) if it doesn't exist. May also fail if the user does not have
        #     # write access to the project.
        #     project = dxpy.DXProject(project_id)
        #     project.new_folder(folder, parents=True)

        build_workflow = kwargs.get("force", False)
        workflow_id = None

        if not build_workflow:
            existing_workflow = list(dxpy.find_data_objects(
                classname="workflow",
                name=workflow_name,
                project=project_id,
                folder=folder,
                describe={
                    "created": True
                }
            ))

            if not existing_workflow:
                build_workflow = True
            else:
                created = existing_workflow[0]["describe"]["created"]
                if wdl_path.stat().st_mtime > created:
                    build_workflow = True
                elif self._import_dirs:
                    for import_dir in self._import_dirs:
                        for imp in import_dir.glob("*.wdl"):
                            if imp.stat().st_mtime > created:
                                build_workflow = True
                                break
                    else:
                        workflow_id = existing_workflow[0]["id"]

        if build_workflow:
            java_args = kwargs.get("java_args", self.java_args) or ""
            imports_args = " ".join(f"-imports {d}" for d in self._import_dirs)
            extras = kwargs.get("extras")
            extras_arg = f"-extras {extras}" if extras else ""
            archive = kwargs.get("archive")
            archive_arg = "-a" if archive else "-f"

            cmd = (
                f"{self.java_bin} {java_args} -jar {self._dxwdl_jar_file} compile "
                f"{wdl_path} -destination {project_id}:{folder} {imports_args} "
                f"{extras_arg} {archive_arg}"
            )

            LOG.info(f"Building workflow with command '{cmd}'")

            try:
                workflow_id = subby.sub(cmd).splitlines(False)[-1]
            except subby.core.CalledProcessError as perr:
                raise ExecutorError(
                    "dxwdl",
                    f"Error building DNAnexus workflow with dxWDL; "
                    f"stdout={perr.stdout}; stderr={perr.stderr}"
                ) from perr

        workflow = dxpy.DXWorkflow(workflow_id)
        DxWdlExecutor._workflow_cache[wdl_path] = workflow
        return workflow

    @classmethod
    def _get_failed_task(cls, analysis: dxpy.DXAnalysis) -> dict:
        """
        Find the causal failure within an execution tree and get the logs.
        """
        query = {
            "project": dxpy.WORKSPACE_ID,
            "state": "failed",
            "describe": {
                "stateTransitions": True
            },
            "include_subjobs": True,
            "root_execution": analysis.get_id()
        }
        cause = None
        cause_ts = None
        for execution_result in dxpy.find_executions(**query):
            if "stateTransitions" in execution_result["describe"]:
                for st in execution_result["describe"]["stateTransitions"]:
                    if st["newState"] == "failed":
                        ts = st["setAt"]
                        if cause is None or ts < cause_ts:
                            cause = execution_result
                            cause_ts = ts

        if cause:
            stdout, stderr = cls._get_logs(cause["id"])
            return {
                "failed_task": cause["id"],
                "failed_task_stdout": stdout,
                "failed_task_stderr": stderr
            }
        else:
            return {
                "msg": f"Analysis {analysis.get_id()} failed but the cause could not "
                       f"be determined"
            }

    @classmethod
    def _get_logs(cls, job_id) -> Tuple[str, ...]:
        logs = ([], [])

        def callback(msg_dict):
            for src, log in zip((STDOUT_LOG, STDERR_LOG), logs):
                if msg_dict["source"] == src:
                    log.append(msg_dict["msg"])
                    break

        client = DXJobLogStreamClient(job_id, msg_callback=callback)
        client.connect()

        return tuple("\n".join(log) for log in logs)

    def _get_analysis_outputs(
        self,
        analysis: dxpy.DXAnalysis,
        expected: Iterable[str],
        default_namespace: str = OUTPUT_STAGE,
    ) -> dict:
        all_outputs = analysis.describe()["output"]
        output = {}

        for key in expected:
            exp_key = key
            if exp_key not in all_outputs and "." not in exp_key:
                exp_key = f"{default_namespace}.{exp_key}"
            if exp_key not in all_outputs:
                raise ValueError(
                    f"Did not find key {exp_key} in outputs of analysis "
                    f"{analysis.get_id()}"
                )
            output[exp_key] = self._resolve_output(all_outputs[exp_key])

        return output

    def _resolve_output(self, value):
        if dxpy.is_dxlink(value):
            dxfile = dxpy.DXFile(value)
            file_id = dxfile.get_id()
            if file_id not in DxWdlExecutor._data_cache:
                # Store each file in a subdirectory named by it's ID to avoid
                # naming collisions
                cache_dir = self._dxwdl_cache_dir / file_id
                cache_dir.mkdir(parents=True)
                filename = cache_dir / dxfile.describe()["name"]
                dxpy.download_dxfile(dxfile, filename)
                DxWdlExecutor._data_cache[file_id] = filename
            return DxWdlExecutor._data_cache[file_id]
        elif isinstance(value, dict):
            return {
                key: self._resolve_output(val)
                for key, val in cast(dict, value).items()
            }
        elif isinstance(value, Sequence) and not isinstance(value, str):
            return [self._resolve_output(val) for val in cast(Sequence, value)]
        else:
            return value


class DxInputsFormatter:
    def __init__(
        self,
        wdl_doc: Document,
        project_id: str = dxpy.PROJECT_CONTEXT_ID,
        data_project_id: Optional[str] = None,
        folder: str = "/",
        data_folder: Optional[str] = None,
        **_
    ):
        self._project_id = data_project_id or project_id
        self._folder = data_folder or folder
        self._data_file_links = set()
        self._wdl_doc = wdl_doc
        self._wdl_decls = dict(
            (d.name, d.value.type) for d in wdl_doc.workflow.available_inputs
        )

    def format_inputs(
        self, inputs_dict: dict, namespace: Optional[str] = None
    ) -> dict:
        prefix = f"{namespace}." if namespace else ""
        formatted = {}

        for key, value in inputs_dict.items():
            new_key = f"{prefix}{key}"

            formatted[new_key] = self.format_value(value, (key,))[0]

            if self._data_file_links:
                formatted[f"{new_key}{DX_FILES_SUFFIX}"] = list(self._data_file_links)
                self._data_file_links.clear()

        return formatted

    def format_value(
        self, value: Any, path: Tuple[str, ...], nested: bool = False
    ) -> Tuple[Any, bool]:
        """
        Convert a primitive, DataFile, Sequence, or Dict to a JSON-serializable object.
        Currently, arbitrary objects can be serialized by implementing an `as_dict()`
        method, otherwise they are converted to strings.

        Args:
            value: The value to format.
            path: The path to the current value.
            nested: Whether the value is nested within a complex type

        Returns:
            The tuple `(val, is_complex)`, where `val` is serializable value and
            `is_complex` is True if the value is a complex type.
        """
        if hasattr(value, "as_dict"):
            return value.as_dict(), True

        if isinstance(value, DataFile):
            return self._format_data_file(cast(DataFile, value)), False

        if isinstance(value, Sequence) and not isinstance(value, str):
            return (
                self._format_sequence(cast(Sequence, value), path, nested=nested),
                True
            )

        if isinstance(value, dict):
            return self._format_dict(cast(dict, value), path, nested=nested), True

        return value, False

    def _format_sequence(
        self, s: Sequence, path: Tuple[str, ...], nested: bool = False
    ) -> Union[list, dict]:
        formatted_seq = []
        is_complex = False

        for val in s:
            formatted_val, val_is_complex = self.format_value(val, path, nested=True)
            formatted_seq.append(formatted_val)
            if val_is_complex:
                is_complex = True

        if nested or not is_complex:
            return formatted_seq
        else:
            return {DX_DICT_KEY: formatted_seq}

    def _format_dict(
        self, d: dict, path: Tuple[str, ...], nested: bool = False
    ) -> dict:
        if self._is_wdl_type(path, Type.Map):
            formatted_dict = {"keys": [], "values": []}

            for key, val in d.items():
                formatted_dict["keys"].append(key)
                formatted_dict["values"].append(
                    self.format_value(val, path + (key,), nested=True)[0]
                )
        else:
            # pair or struct
            formatted_dict = dict(
                (key, self.format_value(val, path + (key,), nested=True)[0])
                for key, val in d.items()
            )

        if nested:
            return formatted_dict
        else:
            return {DX_DICT_KEY: formatted_dict}

    def _is_wdl_type(self, path: Tuple[str, ...], wdl_type: Type) -> bool:
        wdl_decls = self._wdl_decls
        path_len = len(path)

        for i, key in enumerate(path, 1):
            is_last = i == path_len

            for name, type_ in wdl_decls.items():
                if name in {None, key}:
                    if is_last:
                        return isinstance(type_, wdl_type)

                    if isinstance(type_, Type.Array):
                        type_ = cast(Type.Array, type_).item_type

                    if isinstance(type_, Type.Pair):
                        pair_type = cast(Type.Pair, type_)
                        wdl_decls: Dict = {
                            "left": pair_type.left_type,
                            "right": pair_type.right_type
                        }
                    elif isinstance(type_, Type.Map):
                        type_ = cast(Type.Map, type_).item_type[1]
                        # None matches any key
                        wdl_decls: Dict = {None: type_}
                    elif isinstance(type_, Type.StructInstance):
                        wdl_decls = \
                            self._wdl_doc.struct_typedefs[type_.type_name].members

                    break
            else:
                raise ValueError(
                    f"No input matching key {key} at {path[:i]}"
                )

    def _format_data_file(self, df: DataFile) -> dict:
        if isinstance(df.localizer, UrlLocalizer):
            ul = cast(UrlLocalizer, df.localizer)
            if ul.url.startswith("dx://"):
                return dxpy.dxlink(*ul.url[5:].split(":"))

        file_name = df.local_path.name

        existing_files = list(dxpy.find_data_objects(
            classname="file",
            state="closed",
            name=file_name,
            project=self._project_id,
            folder=self._folder,
            recurse=False
        ))

        if not existing_files:
            # TODO: batch uploads and use dxpy.sugar.transfers.Uploader for
            #  parallelization
            return dxpy.dxlink(dxpy.upload_local_file(
                str(df.path),
                name=file_name,
                project=self._project_id,
                folder=self._folder,
                parents=True,
                wait_on_close=True
            ))
        elif len(existing_files) == 1:
            return dxpy.dxlink(existing_files[0]["id"], self._project_id)
        else:
            raise RuntimeError(
                f"Multiple files with name {file_name} found in "
                f"{self._project_id}:{self._folder}"
            )
